from __future__ import annotations
from abc import abstractmethod
import typing
from typing import Tuple

import cohdl

from cohdl._core._ir import repr as ir
from cohdl._core._primitive_type import _PrimitiveType
from cohdl._core._type_qualifier import (
    Port,
    TypeQualifier,
    Generic,
    RefSpec,
    Offset,
    Slice,
)
from cohdl._core._array import Array
from cohdl._core import _boolean, _Boolean
from cohdl._core._intrinsic import _SensitivitySpec, _SensitivityAll, _SensitivityList
from cohdl import (
    Signal,
    Variable,
    Temporary,
    Bit,
    BitOrder,
    BitVector,
    Unsigned,
    Signed,
    Integer,
    Null,
    Full,
    _NullFullType,
)

from cohdl._core._context import EntityInfo
from cohdl._core._inline import InlineVhdl
from cohdl._core._inline import InlineCode as _InlineCode

from cohdl._core import enum as cohdl_enum
from cohdl.utility import IdMap, IdSet
from cohdl.utility.code_writer import IndentBlock, TextBlock


def comment_list(comment: None | str | list[str]):
    if comment is None:
        return []
    if isinstance(comment, str):
        comment = [comment]

    return [f"-- {c}" for c in comment]


class Statement:
    def write(self, scope: VhdlScope) -> TextBlock | str:
        # return TextBlock or str without indentation
        # parent statements define indentation when required
        ...


class Expression(Statement):
    def __init__(self, result):
        self.result = result


class CodeBlock(Statement):
    def __init__(self, stmts: list[Statement]):
        self._stmts = stmts

    def empty(self) -> bool:
        if len(self._stmts) == 0:
            return True

        if all(isinstance(stmt, CodeBlock) for stmt in self._stmts):
            return all(block.empty() for block in self._stmts)
        return False

    def write(self, scope: VhdlScope, indent=True) -> TextBlock:
        return TextBlock(
            content=[
                (
                    stmt.write(scope, indent)
                    if isinstance(stmt, CodeBlock)
                    else stmt.write(scope)
                )
                for stmt in self._stmts
            ]
        )


class Value(Expression):
    def __init__(self, value):
        super().__init__(value)

    def write(self, scope: VhdlScope, target_hint=None, constrain=False):
        return scope.format_value(self.result, target_hint, constrain)


class Constant(Value): ...


class Literal(Constant):
    def write(self, scope: VhdlScope, target_hint=None) -> str:
        if target_hint is None:
            return scope.format_literal(self.result)
        return scope.format_cast(
            target_hint, self.result, scope.format_literal(self.result)
        )


class Target(Value):
    """represents a value that is the target of an assignment

    :param Value: Synthesizable value that is the target of the assignment
    """

    def write(self, scope: VhdlScope):
        return scope.format_target(self.result)


class Source(Value): ...


class Nop(Expression):
    def __init__(self):
        super().__init__(None)

    def write(self, scope: VhdlScope):
        return TextBlock([])


class Comment(Statement):
    def __init__(self, lines):
        super().__init__()
        self.lines = lines

    def write(self, scope: VhdlScope):
        return TextBlock([f"-- {line}" for line in self.lines])


class Boolean(Expression):
    def __init__(self, arg: Expression):
        super().__init__(_Boolean())
        self._arg = arg

    def write(self, scope: VhdlScope):
        arg_str = self._arg.write(scope)

        return scope.format_cast(self.result, self._arg.result, arg_str)


class Event(Expression):
    Type = ir.Event.Type

    def __init__(self, value: Source, event_type: Event.Type):
        self._value = value
        self._event_type = event_type
        super().__init__(_Boolean(True))

    def write(self, scope: VhdlScope):
        Type = Event.Type
        event = self._event_type

        if event is Type.RISING:
            return f"rising_edge({self._value.write(scope)})"
        if event is Type.FALLING:
            return f"falling_edge({self._value.write(scope)})"
        if event is Type.BOTH_EDGES:
            return f"rising_edge({self._value.write(scope)} or falling_edge({self._value.write(scope)})"
        if event is Type.HIGH:
            return f"{self._value.write(scope)} = '1'"
        if event is Type.LOW:
            return f"{self._value.write(scope)} = '0'"

        raise AssertionError()


class Compare(Expression):
    Operator = ir.Compare.Operator

    operator_string = {
        Operator.EQ: "=",
        Operator.NE: "/=",
        Operator.GT: ">",
        Operator.LT: "<",
        Operator.GE: ">=",
        Operator.LE: "<=",
    }

    def __init__(
        self,
        op,
        lhs: Expression,
        rhs: Expression,
        result,
    ):
        super().__init__(result)
        self._op = op
        self._lhs = lhs
        self._rhs = rhs

    def write(self, scope: VhdlScope):
        op = Compare.operator_string[self._op]

        return f"({self._lhs.write(scope)} {op} {self._rhs.write(scope)})"


class All(Expression):
    def __init__(self, args, result):
        super().__init__(result)
        self._args = args

    def write(self, scope: VhdlScope) -> str:
        if len(self._args) == 0:
            return "true"

        return scope.format_cast(
            self.result,
            _Boolean(),
            " and ".join([str(Boolean(arg).write(scope)) for arg in self._args]),
        )


class Any(Expression):
    def __init__(self, args, result):
        super().__init__(result)
        self._args = args

    def write(self, scope: VhdlScope) -> str:
        if len(self._args) == 0:
            return "false"

        return scope.format_cast(
            self.result,
            _Boolean(),
            " or ".join([str(Boolean(arg).write(scope)) for arg in self._args]),
        )


class BinOp(Expression):
    Operator = ir.BinOp.Operator

    operator_string = {
        Operator.ADD: "+",
        Operator.SUB: "-",
        Operator.MUL: "*",
        Operator.BIT_AND: "and",
        Operator.BIT_OR: "or",
        Operator.BIT_XOR: "xor",
        Operator.CONCAT: "&",
        Operator.MOD: "mod",
        Operator.FLOOR_DIV: "/",
    }

    def __init__(
        self,
        op: BinOp.Operator,
        lhs: Expression,
        rhs: Expression,
        result,
    ):
        super().__init__(result)
        self._op = op
        self._lhs = lhs
        self._rhs = rhs

    def write(self, scope: VhdlScope):
        if self._op is BinOp.Operator.LSHIFT:
            # cast rhs to integer
            shift = scope.format_cast(
                Integer(), self._rhs.result, self._rhs.write(scope)
            )
            return f"shift_left({self._lhs.write(scope)}, {shift})"

        if self._op is BinOp.Operator.RSHIFT:
            # cast rhs to integer
            shift = scope.format_cast(
                Integer(), self._rhs.result, self._rhs.write(scope)
            )
            return f"shift_right({self._lhs.write(scope)}, {shift})"

        # cohdl treats Signed and Unsigned as subclasses of BitVector
        # and allows all BitVector operations for them.
        # This includes concatenation. Since VHDL is not so permissive
        # we have to explicitly cast the values to BitVector here.
        if self._op is BinOp.Operator.CONCAT:
            if isinstance(TypeQualifier.decay(self._lhs.result), BitVector):
                self._lhs.result = self._lhs.result.bitvector

            if isinstance(TypeQualifier.decay(self._rhs.result), BitVector):
                self._rhs.result = self._rhs.result.bitvector

        op = BinOp.operator_string[self._op]
        return f"({self._lhs.write(scope)}) {op} ({self._rhs.write(scope)})"


class UnaryOp(Expression):
    Operator = ir.UnaryOp.Operator

    operator_string = {
        Operator.INV: "not ",
        Operator.NEG: "-",
    }

    def __init__(self, op, arg: Expression, result):
        super().__init__(result)
        self._op = op
        self._arg = arg

    def write(self, scope: VhdlScope):
        if self._op is UnaryOp.Operator.NOT:
            as_bool = Boolean(self._arg)

            return scope.format_cast(self.result, bool, f"not ({as_bool.write(scope)})")
        else:
            if self._op is self.Operator.BOOL:
                as_bool = Boolean(self._arg)
                return as_bool.write(scope)

            if self._op is self.Operator.ABS:
                return f"abs({self._arg.write(scope)})"

            op = UnaryOp.operator_string[self._op]
            return f"{op}({self._arg.write(scope)})"


class SignalAssignment(Statement):
    def __init__(self, target: Target, source: Expression):
        self._target = target
        self._source = source

    def write(self, scope: VhdlScope) -> str:
        target = self._target.write(scope)
        source = self._source.write(scope)
        return f"{target} <= {scope.format_cast(self._target.result, self._source.result, source)};"


class VariableAssignment(Statement):
    def __init__(
        self,
        target: Target,
        source: Expression,
    ):
        self._target = target
        self._source = source

    def write(self, scope: VhdlScope) -> str:
        target = self._target.write(scope)
        source = self._source.write(scope)
        return f"{target} := {scope.format_cast(self._target.result, self._source.result, source)};"


class If(Statement):
    def __init__(self, test: Expression, body: CodeBlock, orelse: CodeBlock):
        self._test = test
        self._body = body
        self._orelse = orelse

    def write(self, scope: VhdlScope) -> TextBlock:
        if not self._orelse.empty():
            return TextBlock(
                [
                    f"if {self._test.write(scope)} then",
                    IndentBlock(self._body.write(scope)),
                    "else",
                    IndentBlock(self._orelse.write(scope)),
                    "end if;",
                ],
            )

        return TextBlock(
            [
                f"if {self._test.write(scope)} then",
                IndentBlock(self._body.write(scope)),
                "end if;",
            ],
        )


class SelectWith(Statement):
    def __init__(
        self,
        arg: Value,
        branches: list[Tuple[Constant, Value]],
        default,
        target: Target,
    ):
        self._arg = arg
        self._branches = branches
        self._default = default
        self._target = target

    def write(self, scope) -> TextBlock:
        if self._default is None:
            assert len(self._branches) != 0
            separators = "," * (len(self._branches) - 1) + ";"
        else:
            separators = "," * len(self._branches)

        return TextBlock(
            [
                f"with {self._arg.write(scope, constrain=True)} select {self._target.write(scope)} <=",
                IndentBlock(
                    [
                        *[
                            f"{branch[1].write(scope, self._target.result)} when {branch[0].write(scope, self._arg.result)}{sep}"
                            for branch, sep in zip(self._branches, separators)
                        ],
                        *[
                            f"{default.write(scope, self._target.result)} when others;"
                            for default in [self._default]
                            if default is not None
                        ],
                    ],
                ),
            ],
        )


class CaseWhen(Statement):
    def __init__(
        self,
        cond: Expression,
        branches: list[Tuple[Constant, CodeBlock]],
        others: CodeBlock | None = None,
    ):
        self._cond = cond
        self._branches = branches
        self._others = others

    def write(self, scope) -> TextBlock:
        def write_block(block: CodeBlock) -> TextBlock:
            if block.empty():
                return TextBlock("null;")
            return block.write(scope)

        return TextBlock(
            [
                f"case {self._cond.write(scope, constrain=True)} is",
                *[
                    IndentBlock(
                        [
                            f"when {value.write(scope, target_hint=self._cond.result)} =>",
                            IndentBlock(write_block(block)),
                        ]
                    )
                    for value, block in self._branches
                ],
                (
                    IndentBlock(["when others =>", IndentBlock("null;")])
                    if self._others is None
                    else IndentBlock(
                        ["when others =>", IndentBlock(self._others.write(scope))]
                    )
                ),
                "end case;",
            ],
        )


#
# assertations
#


class Assert(Statement):
    def __init__(self, test: Expression, message):
        self._test = test
        self._message = message

    def write(self, scope: VhdlScope) -> str:
        if self._message is None:
            return f"assert {self._test.write(scope)};"
        return f'assert {self._test.write(scope)} report "{self._message}";'


#
#
#


class InlineCode(Expression):
    def __init__(self, inline: _InlineCode, result):
        super().__init__(result)
        self.inline = inline

    def write(self, scope: VhdlScope) -> str:
        def write_content(content):
            result = ""

            for elem in content:
                if isinstance(elem, InlineVhdl.Text):
                    result += elem.text
                elif isinstance(elem, InlineVhdl.Object):
                    if elem.read:
                        result += scope.format_value(elem.obj)
                    else:
                        result += scope.format_target(elem.obj)
                elif isinstance(elem, InlineVhdl.SubCode):
                    found = False
                    for option in elem.options:
                        if isinstance(option, InlineVhdl):
                            result += write_content(option.content)
                            found = True
                            break

                    assert found, "inline code does not contain vhdl option"
                else:
                    raise AssertionError("invalid type")

            return result

        return self.inline.post_process(write_content(self.inline.content))


#
#
# instances
#
#


class VhdlScope:
    class _VectorSliceHint:
        """
        helper type used to format values
        indicates, that a BitVector was created as a slice of
        a Signed/Unsigned value
        """

        def __init__(self, start, stop):
            self.start = start
            self.stop = stop

    class Declaration:
        def __init__(self, obj, is_active=False, name_hint: str | None = None):
            self.obj = obj
            self.active: bool = is_active
            self.name_hint: str | None = name_hint

            # will be filled during setup
            self.name: str = "NOT_SET"

        def use(self):
            self.active = True

    def __init__(self, parent: VhdlScope | None = None):
        self._setup_complete = False
        self._parent = parent
        self._subscopes: list[VhdlScope] = []
        self._declarations: IdMap[typing.Any, VhdlScope.Declaration] = IdMap()
        self._used_names: set[str] = set()

        if parent is not None:
            parent._subscopes.append(self)

    def reserve_name(self, name):
        self._used_names.add(name)

    def declare(self, obj, _is_first=True, name_hint=None, *, _obj_only=False):
        type_declared = _obj_only

        # use declaration if it is required by multiple subscopes
        # (possibly overwritten by used declaration in parent scope)
        if obj in self._declarations:
            self._declarations[obj].use()

            if not _is_first:
                assert not isinstance(
                    obj, (Variable, Temporary)
                ), "variables and temporaries cannot be shared between scopes"
        else:
            if isinstance(obj, (type, Instance)):
                if _obj_only:
                    return

                if (
                    isinstance(obj, type)
                    and issubclass(obj, Array)
                    and issubclass(
                        obj._elemtype_, (cohdl_enum.Enum, cohdl_enum.DynamicEnum, Array)
                    )
                ):
                    self.declare(obj._elemtype_, True)
                    type_declared = True
            elif isinstance(obj, TypeQualifier):
                # type of signal must be declared before signal
                if (
                    issubclass(
                        obj.type, (cohdl_enum.Enum, cohdl_enum.DynamicEnum, Array)
                    )
                    and not _obj_only
                ):
                    self.declare(obj.type, True)

                if isinstance(obj, TypeQualifier) and not _obj_only:
                    for attr in obj._attributes:
                        self.declare(type(attr), name_hint=attr.name)

                type_declared = True
            else:
                # type of literals must be declared before use
                if isinstance(obj, (cohdl_enum.Enum, cohdl_enum.DynamicEnum)):
                    if not _obj_only:
                        self.declare(type(obj), True)
                        type_declared = True

                    # enumerator literals are already declared in declaration of enum type
                    return

                if isinstance(obj, type) and issubclass(obj, cohdl.Attribute):
                    pass
                else:
                    raise AssertionError(f"ignored declaration {obj}")

            # set first scope to active so it will be used if no enclosing scope
            # declares obj
            self._declarations[obj] = VhdlScope.Declaration(obj, _is_first, name_hint)

            if self._parent is not None:
                self._parent.declare(obj, False, name_hint, _obj_only=type_declared)

    def remove_declaration(self, obj):
        if obj in self._declarations:
            del self._declarations[obj]

            # loop in if because declaration can only exist in subscope
            # if it also exists in parent
            for scope in self._subscopes:
                scope.remove_declaration(obj)

    def complete_setup(self):
        assert not self._setup_complete

        declarations: IdMap[int, VhdlScope.Declaration] = IdMap()

        if self._parent is None:
            if self._used_names is None:
                used_names = set()
            else:
                used_names = set(self._used_names)
        else:
            used_names = set(self._parent._used_names) | self._used_names

        for id, decl in self._declarations.items():
            if decl.active:
                declarations[id] = decl
                for scope in self._subscopes:
                    scope.remove_declaration(id)

        self._declarations = declarations

        for id, decl in declarations.items():
            obj = decl.obj

            name: str

            override = None
            fallback = None

            if isinstance(obj, Port):
                override = obj.name()
            elif isinstance(obj, Signal):
                override = obj.name()
                fallback = "sig"
            elif isinstance(obj, Variable):
                override = obj.name()
                fallback = "var"
            elif isinstance(obj, Temporary):
                override = obj.name()
                fallback = "temp"
            elif isinstance(obj, Concurrent):
                fallback = "concurrent"
            elif isinstance(obj, Process):
                fallback = "proc"
            elif isinstance(obj, Entity):
                override = obj.name()
            elif isinstance(obj, Architecture):
                override = obj.name()
            elif isinstance(obj, EntityInst):
                fallback = "inst"
            elif isinstance(obj, type):
                if issubclass(obj, (cohdl_enum.Enum, cohdl_enum.DynamicEnum)):
                    override = obj.__name__
                elif issubclass(obj, Array):
                    fallback = "array_type"
                elif issubclass(obj, cohdl.Attribute):
                    override = obj.name
                else:
                    raise AssertionError(f"invalid obj {obj}")
            else:
                raise AssertionError(f"invalid obj {obj}")

            if override is not None:
                name = override
            elif decl.name_hint is not None:
                name = decl.name_hint
            elif fallback is not None:
                name = fallback
            else:
                raise AssertionError("Internal error, cannot name object")

            # remove leading and trailing underscores
            # since they are not allowed in vhdl
            name = name.strip("_")

            # avoid name collisions by appending counter to names
            if name.lower() in used_names:
                cnt = 1
                base_name = name
                name = base_name + str(cnt)

                while name.lower() in used_names:
                    cnt *= 2
                    name = base_name + str(cnt)

                step = cnt // 2

                while step:
                    name = base_name + str(cnt - step)

                    if name.lower() not in used_names:
                        cnt -= step

                    step //= 2

                name = base_name + str(cnt)

            decl.name = name
            used_names.add(name.lower())

        self._used_names = used_names
        self._setup_complete = True

        for scope in self._subscopes:
            scope.complete_setup()

    #
    # the following functions are valid after setup is complete
    #

    def declarations(self) -> dict:
        return {decl.name: decl.obj for decl in self._declarations.values()}

    def lookup_name(self, obj) -> str:
        if obj in self._declarations:
            return self._declarations[obj].name

        if self._parent is not None:
            return self._parent.lookup_name(obj)

        raise AssertionError(f"object {obj} not declared in scope")

    def dump(self) -> TextBlock:
        local_decl = []

        for decl in self._declarations.values():
            local_decl.append(
                f"active = {decl.active}, obj = {decl.obj}, name = {decl.name}"
            )

        return TextBlock(
            title=str(self),
            content=[*local_decl, *[sub.dump() for sub in self._subscopes]],
        )

    def _format_ref(
        self,
        obj,
        root_name,
        ref_spec: RefSpec | None,
        is_target: bool,
        constrain: bool = False,
    ):
        if isinstance(ref_spec, Offset):
            ref_spec.simplify()
            assert (
                len(ref_spec.base_offset) == 0
            ), "nested runtime variable offsets not implemented"

            result = f"{root_name}({self.format_value(ref_spec.offset, Integer())})"

            primitive_obj = TypeQualifier.decay(obj)
            obj_type = type(primitive_obj)

            if isinstance(primitive_obj, Array):
                result_type = primitive_obj._elemtype_
            else:
                assert isinstance(primitive_obj, BitVector)
                result_type = Bit

            return result, result_type

        if isinstance(ref_spec, Slice):
            ref_spec.simplify()

            assert (
                len(ref_spec.base_offset) == 0
            ), "runtime variable slices not implemented"

            if ref_spec.start >= ref_spec.stop:
                slice_result = f"{root_name}({ref_spec.start} downto {ref_spec.stop})"
            else:
                slice_result = f"{root_name}({ref_spec.start} to {ref_spec.stop})"

            width = abs(ref_spec.start - ref_spec.stop) + 1

            if is_target:
                return slice_result, BitVector[width]

            obj_type = obj.type

            if constrain:
                if issubclass(obj_type, Unsigned):
                    return f"unsigned'({slice_result})", Unsigned[width]
                if issubclass(obj_type, Signed):
                    return f"signed'({slice_result})", Signed[width]
                return f"std_logic_vector'({slice_result})", BitVector[width]

            if issubclass(obj_type, Unsigned):
                return f"unsigned({slice_result})", Unsigned[width]
            if issubclass(obj_type, Signed):
                return f"signed({slice_result})", Signed[width]
            return f"std_logic_vector({slice_result})", BitVector[width]

        raise AssertionError("error, cannot format input")

    def format_literal(self, obj) -> str:
        if isinstance(obj, (cohdl_enum.Enum, cohdl_enum.DynamicEnum)):
            return obj.name

        if isinstance(obj, bool):
            return str(obj).lower()
        if isinstance(obj, _boolean._Boolean):
            return str(obj._value).lower()
        if isinstance(obj, int):
            return str(obj)
        if isinstance(obj, str):
            return f'"{obj}"'

        if isinstance(obj, (Bit, cohdl.BitState)):
            return f"'{obj}'"
        if isinstance(obj, BitVector):
            if isinstance(obj, Unsigned):
                return f"unsigned'({self.format_literal(obj.bitvector)})"
            if isinstance(obj, Signed):
                return f"signed'({self.format_literal(obj.bitvector)})"
            return f'"{obj}"'
        if isinstance(obj, Integer):
            return f"{obj}"

        # return dummy strings for null and full values
        # replaced with actual values during cast to target type
        if obj is Null:
            return ">>NULL_LITERAL<<"
        if obj is Full:
            return ">>FULL_LITERAL>>"

        if not isinstance(obj, Array):
            assert isinstance(obj, Generic)

            obj = obj._wrapped

            if isinstance(obj, (int, float)):
                return str(obj)
            elif isinstance(obj, str):
                return f'"{obj}"'
        else:
            val = obj._value
            assert val is not None, "array has not default value"
            elemtype = obj._elemtype_

            # use ( 0 => ELEM0, 1 => ELEM1 ) notation because
            # ( ELEM0, ELEM1 ) form is not allowed for arrays with only a single element
            elemstr = [
                f"{nr} => {self.format_literal(elemtype(elem))}"
                for nr, elem in enumerate(val)
            ]

            if len(val) < obj._count_:
                elemstr.append(f"others => {self.format_literal(elemtype())}")

            return f'( {", ".join(elemstr)} )'

        raise AssertionError(f"cannot format {obj}")

    def format_declaration(self, name, obj) -> str:
        if isinstance(self, ProcessScope):
            is_signal = isinstance(obj, Signal)
            is_variable = isinstance(obj, (Variable, Temporary))
        else:
            is_signal = isinstance(obj, (Signal, Temporary))
            is_variable = False

        if is_signal or is_variable:
            type_str = "signal" if is_signal else "variable"
            attributes = []
            for attr in obj._attributes:
                attributes.append(
                    f"attribute {attr.name} of {name} : {type_str} is {self.format_literal(attr.value)};"
                )

        if isinstance(obj, type):
            if issubclass(obj, cohdl_enum.Enum):
                enumerators = list(obj.__members__.keys())
                return f"type {name} is ({', '.join(enumerators)});"
            elif issubclass(obj, cohdl_enum.DynamicEnum):
                enumerators = [member.name for member in obj.__members__]
                return f"type {name} is ({', '.join(enumerators)});"
            elif issubclass(obj, cohdl.Array):
                elemtype = obj._elemtype_

                if issubclass(elemtype, cohdl_enum.Enum):
                    first, *rest = elemtype._member_map_.values()
                    elem_obj = elemtype(first)
                else:
                    elem_obj = elemtype()

                return f"type {name} is array({0} to {obj._count_-1}) of {self.format_type(elem_obj)};"
            elif issubclass(obj, cohdl.Attribute):
                return f"attribute {obj.name} : {self.format_type(obj.attr_type)};"
            else:
                raise AssertionError(f"invalid type {obj}")

        elif is_signal:
            if not obj.has_default():
                return [f"signal {name} : {self.format_type(obj)};", *attributes]
            else:
                return [
                    f"signal {name} : {self.format_type(obj)} := {self.format_literal(obj.default())};",
                    *attributes,
                ]
        elif is_variable:
            assert isinstance(
                self, ProcessScope
            ), "variables can only be used in process scope"
            if not obj.has_default():
                return [f"variable {name} : {self.format_type(obj)};", *attributes]
            else:
                return [
                    f"variable {name} : {self.format_type(obj)} := {self.format_literal(obj.default())};",
                    *attributes,
                ]
        elif isinstance(obj, cohdl.Constant):
            return f"constant {name} : {self.format_type(obj)} := {self.format_literal(obj.get())};"

        raise AssertionError(f"invalid declaration {obj}")

    def format_declarations(self) -> list[str]:
        result = []

        for name, obj in self.declarations().items():
            if not isinstance(obj, Instance):
                decl = self.format_declaration(name, obj)

                if isinstance(decl, str):
                    result.append(decl)
                else:
                    result.extend(decl)

        return result

    def format_type(self, obj) -> str:
        if isinstance(obj, TypeQualifier):
            obj = obj.get()

        if isinstance(obj, (_boolean._Boolean)) or obj is bool:
            return "boolean"
        if isinstance(obj, Bit):
            return "std_logic"
        elif isinstance(obj, BitVector):
            if obj.order is BitOrder.DOWNTO:
                range_ = f"({obj.width - 1} downto 0)"
            else:
                range_ = f"(0 to {obj.width() - 1})"

            if isinstance(obj, Unsigned):
                type_ = "unsigned"
            elif isinstance(obj, Signed):
                type_ = "signed"
            elif isinstance(obj, BitVector):
                type_ = "std_logic_vector"
            else:
                raise AssertionError()

            return f"{type_}{range_}"
        elif isinstance(obj, (cohdl_enum.Enum, cohdl_enum.DynamicEnum)):
            return self.lookup_name(type(obj))
        elif isinstance(obj, cohdl.Array):
            return self.lookup_name(type(obj))
        elif isinstance(obj, Integer):
            assert obj.get_value() is not None
            return "integer"

        if issubclass(obj, str):
            return "string"
        if issubclass(obj, bool):
            return "boolean"
        if issubclass(obj, int):
            return "integer"

        raise AssertionError(f"cannot format {obj}")

    def format_value(self, obj, target_hint=None, constrain=False):
        if isinstance(obj, (cohdl_enum.Enum, cohdl_enum.DynamicEnum)):
            result = obj.name
        elif isinstance(obj, _PrimitiveType):
            result = self.format_literal(obj)
        elif isinstance(obj, (bool, _boolean._Boolean)):
            result = self.format_literal(obj)
        elif isinstance(obj, int):
            result = self.format_literal(obj)
        elif isinstance(obj, str):
            result = self.format_literal(obj)
        elif isinstance(obj, _NullFullType):
            result = self.format_literal(obj)
        elif isinstance(obj, cohdl.BitState):
            result = self.format_literal(obj)
        else:
            assert isinstance(obj, TypeQualifier)
            root_name = self.lookup_name(obj._root)

            if len(obj._ref_spec) == 0:
                result = self.format_vhdl_cast(obj, root_name)
            else:
                result = root_name
                parent = obj._root

                for ref in obj._ref_spec:
                    result, result_type = self._format_ref(
                        parent, result, ref, False, constrain
                    )
                    parent = ref.obj

                result = self.format_cast(obj.copy(), Signal[result_type](), result)

        if target_hint is None:
            return result

        return self.format_cast(target_hint, obj, result)

    def format_target(self, obj):
        assert isinstance(obj, (Signal, Variable, Temporary))

        root = obj._root
        result = self.lookup_name(root)
        parent = root

        for ref in obj._ref_spec:
            result, result_type = self._format_ref(parent, result, ref, True)

        return result

    def format_vhdl_cast(self, value, value_str):
        """
        perform cast so the vhdl type matches to cohdl type
        (required for bitvector/signed/unsigned properties)
        """

        if not isinstance(value, TypeQualifier):
            return value_str

        if not issubclass(value.type, BitVector):
            return value_str

        root_type = value._root.type

        if issubclass(root_type, Array):
            root_type = root_type.elemtype()

        value_type = value.type

        if root_type is value.type:
            return value_str

        if root_type.width == value_type.width:
            vhdl_type = root_type
        else:
            vhdl_type = BitVector[value_type.width]

        if issubclass(vhdl_type, Unsigned):
            if issubclass(value_type, Signed):
                return f"signed(std_logic_vector({value_str}))"
            if issubclass(value_type, BitVector):
                return f"std_logic_vector({value_str})"

        if issubclass(vhdl_type, Signed):
            if issubclass(value_type, Unsigned):
                return f"unsigned(std_logic_vector({value_str}))"
            if issubclass(value_type, BitVector):
                return f"std_logic_vector({value_str})"

        assert issubclass(vhdl_type, BitVector)

        if issubclass(value_type, Unsigned):
            return f"unsigned({value_str})"

        if issubclass(value_type, Signed):
            return f"signed({value_str})"

        raise AssertionError(f"assignment not supported {vhdl_type} -> {value.type}")

    def format_cast(self, target, value, value_str):
        """
        wrap value_str that corresponds to value in a cast function
        to make it assignable to target
        """

        target_type = type(TypeQualifier.decay(target))
        value_type = type(TypeQualifier.decay(value))

        vhdl_target_type = target_type

        if isinstance(target, TypeQualifier):
            if issubclass(target_type, BitVector):
                vhdl_target_type = type(TypeQualifier.decay(target._root))

                if target._ref_spec is not None:
                    for ref_spec in target._ref_spec:
                        if issubclass(vhdl_target_type, Array):
                            if isinstance(ref_spec, Offset):
                                vhdl_target_type = vhdl_target_type._elemtype_
                            else:
                                raise AssertionError("Array slices are not implemented")
                        else:
                            if issubclass(vhdl_target_type, Signed):
                                vhdl_target_type = Signed[target_type.width]
                            elif issubclass(vhdl_target_type, Unsigned):
                                vhdl_target_type = Unsigned[target_type.width]
                            else:
                                vhdl_target_type = BitVector[target_type.width]

        if issubclass(target_type, Bit):
            if issubclass(value_type, Bit):
                return value_str
            if issubclass(value_type, _Boolean):
                return f"cohdl_bool_to_std_logic({value_str})"
            if issubclass(value_type, _NullFullType):
                return self.format_literal(target_type(value))

        elif issubclass(target_type, (_Boolean, bool)):
            if issubclass(value_type, (_Boolean, bool)):
                return value_str
            if issubclass(value_type, Bit):
                return f"{value_str} = '1'"
            if issubclass(value_type, (Unsigned, Signed)):
                return f"({value_str} /= 0)"
            if issubclass(value_type, BitVector):
                zeros = "0" * value_type.width
                return f'({value_str} /= "{zeros}")'
            if issubclass(value_type, Integer):
                return f"({value_str} /= 0)"

        elif issubclass(target_type, BitVector):
            if issubclass(value_type, _NullFullType) or isinstance(value, BitVector):
                return self.format_literal(vhdl_target_type(value))

            if issubclass(value_type, Integer):
                if issubclass(target_type, Unsigned):
                    value_str = f"to_unsigned({value_str}, {target_type.width})"

                    if issubclass(vhdl_target_type, Unsigned):
                        return value_str
                    elif issubclass(vhdl_target_type, Signed):
                        return f"signed(std_logic_vector({value_str}))"
                    else:
                        return f"std_logic_vector({value_str})"

                elif issubclass(target_type, Signed):
                    value_str = f"to_signed({value_str}, {target_type.width})"

                    if issubclass(vhdl_target_type, Signed):
                        return value_str
                    elif issubclass(vhdl_target_type, Unsigned):
                        return f"unsigned(std_logic_vector({value_str}))"
                    else:
                        return f"std_logic_vector({value_str})"

                raise AssertionError("invalid target for integer assignment")

            if issubclass(vhdl_target_type, Unsigned):
                if issubclass(value_type, Unsigned):
                    if target_type.width == value_type.width:
                        return value_str
                    else:
                        assert target_type.width > value_type.width
                        return f"resize({value_str}, {target_type.width})"
                elif issubclass(value_type, Signed):
                    if issubclass(target_type, Signed):
                        if target_type.width != value_type.width:
                            assert target_type.width > value_type.width
                            value_str = f"resize({value_str}, {target_type.width})"
                    else:
                        assert target_type.width == value_type.width

                    return f"unsigned(std_logic_vector({value_str}))"
                else:
                    assert issubclass(value_type, BitVector)
                    assert target_type.width == value_type.width
                    return f"unsigned({value_str})"

            if issubclass(vhdl_target_type, Signed):
                if issubclass(value_type, Signed):
                    if target_type.width == value_type.width:
                        return value_str
                    else:
                        assert target_type.width > value_type.width
                        return f"resize({value_str}, {target_type.width})"
                elif issubclass(value_type, Unsigned):
                    if issubclass(target_type, Unsigned):
                        if target_type.width != value_type.width:
                            assert target_type.width > value_type.width
                            value_str = f"resize({value_str}, {target_type.width})"
                    elif issubclass(target_type, Signed):
                        assert target_type.width >= value_type.width
                        value_str = f"resize({value_str}, {target_type.width})"
                    else:
                        assert target_type.width == value_type.width

                    return f"signed(std_logic_vector({value_str}))"
                else:
                    assert issubclass(value_type, BitVector)
                    assert target_type.width == value_type.width
                    return f"signed({value_str})"

            if issubclass(vhdl_target_type, BitVector):
                if issubclass(value_type, Signed):
                    if issubclass(target_type, Signed):
                        if target_type.width != value_type.width:
                            assert target_type.width > value_type.width
                            value_str = f"resize({value_str}, {target_type.width})"
                    else:
                        assert not issubclass(target_type, Unsigned)
                        assert target_type.width == value_type.width

                    return f"std_logic_vector({value_str})"

                elif issubclass(value_type, Unsigned):
                    if issubclass(target_type, Unsigned):
                        if target_type.width != value_type.width:
                            assert target_type.width > value_type.width
                            value_str = f"resize({value_str}, {target_type.width})"
                    elif issubclass(target_type, Signed):
                        assert target_type.width > value_type.width
                        value_str = f"resize({value_str}, {target_type.width})"
                    else:
                        assert target_type.width == value_type.width

                    return f"std_logic_vector({value_str})"

                else:
                    assert issubclass(value_type, BitVector)
                    assert target_type.width == value_type.width
                    return value_str

            raise AssertionError(
                f"invalid {vhdl_target_type} - {target_type} - {value_type}"
            )

        elif issubclass(target_type, Integer):
            if issubclass(value_type, (Integer, int)):
                return value_str
            if issubclass(value_type, (Signed, Unsigned)):
                return f"to_integer({value_str})"

        elif issubclass(target_type, (cohdl_enum.Enum, cohdl_enum.DynamicEnum)):
            assert target_type is value_type
            return value_str

        elif issubclass(target_type, Array):
            assert issubclass(value_type, Array)
            assert target_type._elemtype_ is value_type._elemtype_
            return value_str

        raise AssertionError(f"error cannot convert from {value_type} to {target_type}")


class ProcessScope(VhdlScope):
    def declare(self, obj, _is_first=True, name_hint=None):
        # signals can't be declared in process scope
        if isinstance(obj, Signal) or isinstance(obj, Instance):
            assert self._parent is not None
            self._parent.declare(obj, True, name_hint)
        else:
            super().declare(obj, _is_first, name_hint)


class ModuleScope(VhdlScope):
    # source: https://redirect.cs.umbc.edu/portal/help/VHDL/reserved.html
    _vhdl_reserved = {
        "abs",
        "access",
        "after",
        "alias",
        "all",
        "and",
        "architecture",
        "array",
        "assert",
        "attribute",
        "begin",
        "block",
        "body",
        "buffer",
        "bus",
        "case",
        "component",
        "configuration",
        "constant",
        "disconnect",
        "downto",
        "else",
        "elsif",
        "end",
        "entity",
        "exit",
        "file",
        "for",
        "function",
        "generate",
        "generic",
        "group",
        "guarded",
        "if",
        "impure",
        "in",
        "inertial",
        "inout",
        "is",
        "label",
        "library",
        "linkage",
        "literal",
        "loop",
        "map",
        "mod",
        "nand",
        "new",
        "next",
        "nor",
        "not",
        "null",
        "of",
        "on",
        "open",
        "or",
        "others",
        "out",
        "package",
        "port",
        "postponed",
        "procedure",
        "process",
        "pure",
        "range",
        "record",
        "register",
        "reject",
        "rem",
        "report",
        "return",
        "rol",
        "ror",
        "select",
        "severity",
        "signal",
        "shared",
        "sla",
        "sll",
        "sra",
        "srl",
        "subtype",
        "then",
        "to",
        "transport",
        "type",
        "unaffected",
        "units",
        "until",
        "use",
        "variable",
        "wait",
        "when",
        "while",
        "with",
        "xnor",
        "xor",
        "default",
    }

    _additional_reserved = {
        "std_logic",
        "std_logic_vector",
        "signed",
        "unsigned",
        "resize",
    }

    def __init__(self):
        super().__init__()

        self._used_names = self._vhdl_reserved | self._additional_reserved


class EntityScope(VhdlScope): ...


class ArchScope(VhdlScope): ...


class AliasScope:
    def __init__(self, baseObject):
        self.__class__ = type(
            baseObject.__class__.__name__, (self.__class__, baseObject.__class__), {}
        )

        self.__dict__ = baseObject.__dict__
        self.__class__._alias_map_ = IdMap()  # type: ignore

    def set_alias(self, obj, replacement):
        self.__class__._alias_map_[obj] = replacement  # type: ignore

    def lookup_name(self, obj) -> str:
        if obj in self.__class__._alias_map_:  # type: ignore
            obj = self.__class__._alias_map_[obj]  # type: ignore
        return super().lookup_name(obj)  # type: ignore


class Instance:
    def __init__(self, scope: VhdlScope):
        self._scope = scope

    def scope(self):
        return self._scope

    @abstractmethod
    def write(self): ...

    def dump(self):
        return self.write()

    def sub_entities(self) -> IdSet[Entity]:
        return IdSet()


class Entity(Instance):
    def __init__(
        self,
        info: EntityInfo,
        parent_scope: VhdlScope,
        instances: list[Instance],
    ):
        super().__init__(EntityScope(parent_scope))
        parent_scope.declare(self)

        attributes = info.attributes

        sub_entities = IdSet()
        sub_entities.update(
            [
                inst
                for inst in instances
                if isinstance(inst, EntityInst) and not inst.extern()
            ]
        )

        self._name = info.name
        self._instances: list[Instance] = instances
        self._ports = info.ports
        self._generics = info.generics
        self._sub_entities = sub_entities
        self._path = "work" if "path" not in attributes else attributes["path"]
        self._extern = info.extern

        if info.extern:
            self._arch_name = attributes.get("arch_name", None)
        else:
            self._arch_name = attributes.get("arch_name", f"arch_{info.name}")

        self._arch: Architecture | None = None

    def architecture(self):
        assert self._arch is not None
        return self._arch

    def name(self):
        return self._name

    def path(self):
        return self._path

    def extern(self):
        return self._extern

    def sub_entities(self):
        return self._sub_entities

    def ports(self) -> dict[str, Port]:
        return self._ports

    def add_block(self, block: Block):
        self._blocks.append(block)

    def _generic_declarations(self) -> list[str]:
        raise AssertionError(
            "not implemented, generics can only be used when instantiating external entities"
        )

    def _port_declarations(self) -> list[str]:
        ret = []

        for name, port in self._ports.items():
            direction = port.direction()
            obj = port.get()

            if direction.is_input():
                dir_str = "in"
            elif direction.is_output():
                dir_str = "out"
            elif direction.is_inout():
                dir_str = "inout"
            else:
                raise AssertionError("invalid direction")

            ret.append(f"{name} : {dir_str} {self._scope.format_type(obj)};")

        if len(ret) != 0:
            # remove terminating semicolon
            ret[-1] = ret[-1][:-1]

        return ret

    def _generic_map(self) -> TextBlock:
        return TextBlock(
            title="generic (", content=[self._generic_declarations(), ");"]
        )

    def _port_map(self) -> TextBlock:
        return TextBlock(title="port (", content=[self._port_declarations(), ");"])

    def _library_declaration(self) -> TextBlock:
        extern_libraries = set()

        for entity in self._sub_entities:
            assert isinstance(entity, EntityInst)
            path = entity._entity.path()
            if path is not None and path != "work":
                lib_name = path.split(".")[0]
                extern_libraries.update([f"library {lib_name.lower()};"])

        return TextBlock(
            [
                "library ieee;",
                "use ieee.std_logic_1164.all;",
                "use ieee.numeric_std.all;",
                *extern_libraries,
            ]
        )

    def _entity_declaration(self) -> TextBlock:
        return TextBlock(
            [
                f"entity {self._name} is",
                IndentBlock(self._port_map()),
                f"end {self._name};",
            ]
        )

    def write(self) -> TextBlock:
        assert self._arch is not None

        return TextBlock(
            [
                self._library_declaration(),
                "\n",
                self._entity_declaration(),
                "\n",
                self._arch.write(),
            ],
        ).dump()


class Architecture(Instance):
    def __init__(self, scope: ArchScope, entity: Entity, instances: list[Instance]):
        super().__init__(scope)
        self._entity = entity
        self._instances = instances

    def name(self) -> str:
        hint = self._entity._arch_name

        if hint is None:
            return f"arch_{self._entity.name()}"
        return hint

    def arch_name(self):
        return self._scope.lookup_name(self)

    def entity_name(self):
        return self._scope.lookup_name(self._entity)

    def write_declarations(self):
        return self._scope.format_declarations()

    def write_instances(self):
        return [inst.write() for inst in self._instances]

    def write(self) -> TextBlock:
        return TextBlock(
            [
                f"architecture {self.arch_name()} of {self.entity_name()} is",
                # TODO cleanup
                IndentBlock(
                    [
                        *[
                            "function cohdl_bool_to_std_logic(inp: boolean) return std_logic is",
                            "begin",
                            "  if inp then",
                            "    return('1');",
                            "  else",
                            "    return('0');",
                            "  end if;",
                            "end function cohdl_bool_to_std_logic;",
                        ],
                        *self.write_declarations(),
                    ],
                ),
                "begin",
                IndentBlock(self.write_instances()),
                f"end architecture {self.arch_name()};",
            ]
        )


class Block(Instance):
    def __init__(
        self,
        parent_scope: VhdlScope,
        subblocks: list[Instance],
        name: str | None,
        attributes: dict,
    ):
        # use parent scope because declaration at block level is not supported
        super().__init__(parent_scope)

        self._subblocks = subblocks
        self._name = name
        self._attributes = attributes

    def sub_entities(self) -> IdSet[Entity]:
        return IdSet.union(*[subblock.sub_entities() for subblock in self._subblocks])

    def write(self) -> TextBlock:
        return TextBlock(
            [
                *([f"-- Block ({self._name})"] if self._name is not None else []),
                *comment_list(self._attributes.get("comment", None)),
                *[subblock.write() for subblock in self._subblocks],
                "\n",
            ],
        )


class Concurrent(Instance):
    def __init__(
        self,
        parent_scope: VhdlScope,
        stmts: list[Statement],
        name: str,
        attributes: dict,
    ):
        # use parent scope because declaration at concurrent level is not supported
        super().__init__(parent_scope)

        self._stmts = stmts
        self._name = name
        self._attributes = attributes

    def write(self) -> TextBlock:
        if self._name is None:
            return TextBlock(
                [
                    *comment_list(self._attributes.get("comment", None)),
                    *[stmt.write(self._scope) for stmt in self._stmts],
                ],
            )

        return TextBlock(
            [
                f"",
                f"-- CONCURRENT BLOCK ({self._name})",
                *comment_list(self._attributes.get("comment", None)),
                *[stmt.write(self._scope) for stmt in self._stmts],
            ],
        )


class Process(Instance):
    def __init__(
        self,
        scope: ProcessScope,
        code: CodeBlock,
        sensitivity: _SensitivitySpec,
        attributes: dict,
    ):
        super().__init__(scope)

        self._sensitivity = sensitivity
        self._code = code
        self._attributes = attributes

    def _write_declarations(self) -> list[str]:
        return self._scope.format_declarations()

    def _write_header(self) -> TextBlock:
        if isinstance(self._sensitivity, _SensitivityAll):
            sensitivity_list = "all"
        else:
            assert isinstance(self._sensitivity, _SensitivityList)
            sensitivity_list = ", ".join(
                self._scope.format_value(item) for item in self._sensitivity.signals
            )

        proc_name = self._scope.lookup_name(self)

        return TextBlock(
            content=[
                f"{proc_name}: process({sensitivity_list})",
                IndentBlock(self._write_declarations()),
            ],
        )

    def _write_body(self) -> TextBlock:
        return TextBlock(["begin", IndentBlock(self._code.write(self._scope, True))])

    def _write_end(self) -> str:
        return f"end process;"

    def write(self) -> TextBlock:
        return TextBlock(
            [
                "\n",
                *comment_list(self._attributes.get("comment", None)),
                self._write_header(),
                self._write_body(),
                self._write_end(),
            ],
        )


class EntityInst(Instance):
    def __init__(
        self,
        parent_scope: VhdlScope,
        entity: Entity,
        ports: dict[str, Signal],
        generics: dict[str, Any],
    ):
        super().__init__(parent_scope)

        parent_scope.declare(self, name_hint=f"comp_{entity.name()}")

        self._entity = entity
        self._ports = ports
        self._generics = generics

    def extern(self) -> bool:
        return self._entity.extern()

    def sub_entities(self) -> IdSet[Entity]:
        if self.extern():
            return IdSet()

        result = IdSet.union(self._entity.sub_entities())
        result.update([self._entity])
        return result

    def _generic_map(self) -> list[str]:
        entity = self._entity
        if len(entity._generics) == 0:
            return []

        generic_map: list[Tuple[str, str]] = []

        for generic_name in entity._generics:
            generic_map.append(
                (
                    generic_name,
                    self._scope.format_value(
                        self._generics[generic_name],
                    ),
                )
            )

        line_end = [","] * (len(generic_map) - 1) + [""]

        return [
            "generic map(",
            *[
                f"{port_name} => {local}{sep}"
                for (port_name, local), sep in zip(generic_map, line_end)
            ],
            ");",
        ]

    def _port_map(self) -> list[str]:
        if len(self._ports) == 0:
            return []

        port_map: list[Tuple[str, str]] = []

        for port_name in self._entity.ports():
            port_map.append(
                (port_name, self._scope.format_target(self._ports[port_name]))
            )

        line_end = [","] * (len(port_map) - 1) + [""]

        return [
            "port map(",
            *[
                f"{port_name} => {local}{sep}"
                for (port_name, local), sep in zip(port_map, line_end)
            ],
            ");",
        ]

    def write(self) -> TextBlock:
        port_map: list[Tuple[str, str]] = []

        for port_name in self._entity.ports():
            port_map.append(
                (port_name, self._scope.format_value(self._ports[port_name]))
            )

        entity_name = self._entity._name
        arch_name = self._entity._arch_name
        arch_spec = "" if arch_name is None else f"({arch_name})"
        path = self._entity._path

        comp_name = self._scope.lookup_name(self)

        return TextBlock(
            title=f"{comp_name}: entity {path}.{entity_name}{arch_spec}",
            content=[*self._generic_map(), *self._port_map()],
        )


import os


class Library(Instance):
    @staticmethod
    def from_top_entity(top_entity: Entity):
        entities = IdSet()

        def collect_subenties(parent_entity):
            for entity_inst in parent_entity.sub_entities():
                entity = entity_inst._entity
                entities.add(entity)
                collect_subenties(entity)

        collect_subenties(top_entity)

        return Library(
            top_entity,
            [top_entity, *entities],
            parent_scope=VhdlScope(),
        )

    def __init__(self, top_entity: Entity, entities: list[Entity], parent_scope):
        super().__init__(parent_scope)
        self._top_entity = top_entity
        self._entities: list[Entity] = entities

    def top_entity(self):
        return self._top_entity

    def write(self):
        return TextBlock([entity.write() for entity in self._entities]).dump()

    def write_dir(self, path):
        file_list = []

        for entity in self._entities:
            file_path = os.path.join(path, f"{entity.name()}.vhd")
            file_list.append(file_path)

            with open(file_path, "w") as file:
                print(entity.write(), file=file)

        return file_list



from __future__ import annotations

from cohdl import Signed, Unsigned, std, Signal, BitVector, Bit, Null


class MyRecord(std.Record):
    a: Unsigned[7]
    b: Signed[4]
    c: BitVector[43]
    d: Bit


class Derived(MyRecord):
    asdf: Bit


a = std.make(Derived)
b = std.make[Signal](Derived, a)
c = std.make[std.identity](Derived, a)

print()
print(repr(a))
print(std.to_bits(a))

print()
print(repr(b))
print(std.to_bits(b))

print()
print(repr(c))
print(std.to_bits(c))


print("\n\nDONE\n\n")
exit()


class WidthType(int):
    pass


class MyRecordTemplate(std.Record[WidthType]):
    a: Unsigned[WidthType]
    b: Signed[WidthType]


print("\n\nDONE\n\n")

m = MyRecordTemplate[4](Signal[Unsigned[4]](), b=Signed[4]())
print("DONE")
m = MyRecordTemplate[4](a=Signal[Unsigned[4]](), b=Signed[4]())
m = MyRecordTemplate[5](b=Signed[5](12), a=Unsigned[5](7))

m = std.make(MyRecordTemplate[3], Unsigned[3](7), Signed[3](1))
v = std.make(MyRecordTemplate[3], Null, ~Null)

# e = Example[123]()

bits = std.to_bits(m)

print(bits)

n = std.from_bits[MyRecordTemplate[3]](bits)

print(n)
print(v)
print(repr(n))
print(n == n)
print(n != n)



from __future__ import annotations

from typing import Callable

import inspect
import enum

import cohdl
from cohdl import (
    BitSignalEvent,
    Bit,
    Signal,
    evaluated,
    pyeval,
    Variable,
    AssignMode,
)

from cohdl.utility.source_location import SourceLocation

from ._prefix import _Prefix
from ._core_utility import as_awaitable


class Reset:
    def __init__(
        self,
        signal: cohdl.Signal[cohdl.Bit],
        active_low: bool = False,
        is_async: bool = False,
    ):
        assert isinstance(signal, cohdl.Signal)
        assert issubclass(signal.type, cohdl.Bit)

        self._signal = signal
        self._active_low = active_low
        self._is_async = is_async

    def __bool__(self):
        if self._active_low:
            return not self._signal
        else:
            return self._signal

    def is_async(self) -> bool:
        return self._is_async

    def is_active_low(self) -> bool:
        return self._active_low

    def is_active_high(self) -> bool:
        return not self._active_low

    def active_high_signal(self) -> cohdl.Signal[cohdl.Bit]:
        if self._active_low:
            if evaluated():
                return ~self._signal
            else:
                out = cohdl.Signal[cohdl.Bit]()

                @concurrent
                def inv_reset():
                    out.next = ~self._signal

                return out
        else:
            return self._signal

    def active_low_signal(self) -> cohdl.Signal[cohdl.Bit]:
        if not self._active_low:
            if evaluated():
                return ~self._signal
            else:
                out = cohdl.Signal[cohdl.Bit]()

                @concurrent
                def inv_reset():
                    out.next = ~self._signal

                return out
        else:
            return self._signal

    def signal(self) -> cohdl.Signal[cohdl.Bit]:
        return self._signal


class Frequency:
    @staticmethod
    @pyeval
    def _getter(arg, factor):
        if isinstance(arg, Frequency):
            return arg._val / factor
        return Frequency(arg * factor)

    @pyeval
    def __init__(self, val: int | float | Frequency | Duration):
        if isinstance(val, Duration):
            val = val.frequency()
        if isinstance(val, Frequency):
            val = val._val

        self._val = float(val)

    @pyeval
    def frequency(self) -> Frequency:
        return self

    @pyeval
    def period(self) -> Duration:
        return Duration(1 / self._val)

    @pyeval
    def gigahertz(self) -> float:
        return Frequency._getter(self, 1e9)

    @pyeval
    def megahertz(self) -> float:
        return Frequency._getter(self, 1e6)

    @pyeval
    def kilohertz(self) -> float:
        return Frequency._getter(self, 1e3)

    @pyeval
    def hertz(self) -> float:
        return Frequency._getter(self, 1)

    @pyeval
    def __eq__(self, other: Frequency):
        return self._val == other._val


class Duration:
    @staticmethod
    @pyeval
    def _getter(arg, factor):
        if isinstance(arg, Duration):
            return factor / arg._freq._val
        return Duration(arg / factor)

    @pyeval
    def __init__(self, val: int | float | Frequency | Duration):
        if isinstance(val, Duration):
            self._freq = val._freq
        elif isinstance(val, Frequency):
            self._freq = val
        else:
            self._freq = Frequency(1 / val)

    @pyeval
    def frequency(self) -> Frequency:
        return self._freq

    @pyeval
    def period(self) -> Duration:
        return self

    @pyeval
    def picoseconds(self):
        return Duration._getter(self, 1e12)

    @pyeval
    def nanoseconds(self):
        return Duration._getter(self, 1e9)

    @pyeval
    def microseconds(self):
        return Duration._getter(self, 1e6)

    @pyeval
    def milliseconds(self):
        return Duration._getter(self, 1e3)

    @pyeval
    def seconds(self):
        return Duration._getter(self, 1)

    @pyeval
    def __eq__(self, other: Duration):
        return self._freq == other._freq

    @pyeval
    def count_periods(
        self, subperiod: Duration, *, allowed_delta=1e-9, float_result=False
    ):
        self_ps = self.picoseconds()
        other_ps = subperiod.picoseconds()

        real_result = self_ps / other_ps

        if float_result:
            return real_result
        else:
            rounded = round(real_result)
            assert (
                abs((rounded - real_result) / real_result) <= allowed_delta
            ), f"subperiod does not divide period {real_result=} {rounded=}"
            return rounded

    @pyeval
    def __truediv__(self, other: int | float | Duration):
        if isinstance(other, Duration):
            self_ps = self.picoseconds()
            other_ps = other.picoseconds()
            return self_ps / other_ps
        else:
            return type(self).picoseconds(self.picoseconds() / other)


Period = Duration


class ClockEdge(enum.Enum):
    NEITHER = enum.auto()
    RISING = enum.auto()
    FALLING = enum.auto()
    BOTH = enum.auto()

    @pyeval
    def event_type(self) -> BitSignalEvent.Type:
        if self is ClockEdge.RISING:
            return BitSignalEvent.Type.RISING
        elif self is ClockEdge.FALLING:
            return BitSignalEvent.Type.FALLING
        elif self is ClockEdge.BOTH:
            return BitSignalEvent.Type.BOTH_EDGES

        raise AssertionError(f"invalid value {self}")


class Clock:
    Edge = ClockEdge

    def __init__(
        self,
        clk_signal: cohdl.Signal[cohdl.Bit],
        *,
        active_edge: ClockEdge = ClockEdge.RISING,
        frequency: Frequency | int | None = None,
        period: Duration | int | None = None,
        duty: float = 0.5,
        phase: float = 0.0,
    ):
        assert isinstance(clk_signal, cohdl.Signal)
        assert issubclass(clk_signal.type, cohdl.Bit)
        assert (
            frequency is None or period is None
        ), "only one of frequency and period can be set"

        self._signal = clk_signal
        self._duty = duty
        self._phase = phase
        self._edge = active_edge

        if frequency is not None:
            self._frequency = Frequency(frequency)
        elif period is not None:
            self._frequency = Frequency(period)
        else:
            self._frequency = None

    def __bool__(self):
        if self._edge is ClockEdge.RISING:
            return cohdl.rising_edge(self._signal)
        elif self._edge is ClockEdge.FALLING:
            return cohdl.falling_edge(self._signal)
        elif self._edge is ClockEdge.BOTH:
            return cohdl.rising_edge(self._signal) | cohdl.falling_edge(self._signal)
        else:
            raise AssertionError("invalid clock edge")

    def is_rising_edge(self) -> bool:
        return self._edge is ClockEdge.RISING

    def is_falling_edge(self) -> bool:
        return self._edge is ClockEdge.FALLING

    def duration_ticks(self, duration: Duration):
        clk_duration = self._frequency.period()
        return duration // clk_duration

    def edge(self):
        return self._edge

    def signal(self) -> cohdl.Signal[cohdl.Bit]:
        return self._signal

    def frequency(self):
        return self._frequency

    def period(self):
        return self._frequency.period()

    def duty(self):
        return self.duty

    def phase(self):
        return self.phase

    def rising(self) -> Clock:
        return Clock(
            self._signal,
            active_edge=ClockEdge.RISING,
            frequency=self._frequency,
            duty=self._duty,
            phase=self._phase,
        )

    def falling(self) -> Clock:
        return Clock(
            self._signal,
            active_edge=ClockEdge.FALLING,
            frequency=self._frequency,
            duty=self._duty,
            phase=self._phase,
        )

    def both(self) -> Clock:
        return Clock(
            self._signal,
            active_edge=ClockEdge.BOTH,
            frequency=self._frequency,
            duty=self._duty,
            phase=self._phase,
        )


def block(fn: Callable | None = None, /, comment=None, attributes: dict | None = None):
    if attributes is None:
        attributes = {}

    if comment is None:
        assert "comment" not in attributes, "comment attribute already set"
        attributes["comment"] = comment

    if fn is None:

        def helper(function):
            with cohdl._core._context.Block("", attributes=attributes):
                function()

        return helper
    else:
        with cohdl._core._context.Block("", attributes=attributes):
            fn()


class _NopContextManager:
    def __enter__(self):
        return None

    def __exit__(self, type, value, traceback):
        return None


def _prefix_wrapper(prefix, fn):
    def wrapped():
        with prefix:
            fn()

    return wrapped


def concurrent(
    fn: Callable | None = None,
    *,
    comment=None,
    attributes: dict | None = None,
    capture_lazy: bool = False,
    wrapped_fn=None,
):
    parent_prefix = _Prefix._parent_prefix()

    if parent_prefix is None:
        parent_prefix = _NopContextManager()

    if attributes is None:
        attributes = {}

    if comment is not None:
        assert "comment" not in attributes, "comment attribute already set"
        attributes["comment"] = comment

    if fn is None:

        def wrapper(fn):
            nonlocal wrapped_fn
            wrapped_fn = fn if wrapped_fn is None else wrapped_fn
            cohdl.concurrent_context(
                _prefix_wrapper(parent_prefix, fn),
                name=fn.__name__,
                attributes=attributes,
                source_location=SourceLocation.from_function(wrapped_fn),
                captured_functions=None if capture_lazy else [fn],
            )

        return wrapper

    wrapped_fn = fn if wrapped_fn is None else wrapped_fn

    cohdl.concurrent_context(
        _prefix_wrapper(parent_prefix, fn),
        name=fn.__name__,
        attributes=attributes,
        source_location=SourceLocation.from_function(wrapped_fn),
        captured_functions=None if capture_lazy else [fn],
    )


def _sequential_impl(
    trigger=None,
    /,
    reset=None,
    *,
    step_cond=None,
    comment=None,
    attributes: dict | None = None,
    capture_lazy: bool = False,
    wrapped_fn,
    on_reset=None,
):
    if on_reset is None:
        on_reset = []
    elif not isinstance(on_reset, list):
        on_reset = [on_reset]

    if trigger is None:

        def wrapper(fn):
            _sequential_impl(
                fn,
                reset=reset,
                step_cond=step_cond,
                comment=comment,
                attributes=attributes,
                capture_lazy=capture_lazy,
                wrapped_fn=wrapped_fn,
            )

        return wrapper

    parent_prefix = _Prefix._parent_prefix()

    if parent_prefix is None:
        parent_prefix = _NopContextManager()

    is_coro = inspect.iscoroutinefunction(trigger)

    if attributes is None:
        attributes = {}

    if comment is not None:
        assert "comment" not in attributes, "comment attribute already set"
        attributes["comment"] = comment

    if step_cond is None:
        step_cond = lambda: True

    if inspect.isfunction(trigger) or is_coro:
        wrapped_fn = trigger if wrapped_fn is None else wrapped_fn

        if is_coro:
            callable = coro = trigger()
            fn = None
        else:
            callable = fn = trigger
            coro = None

        def wrapper():
            cohdl.sensitivity.all()
            cohdl.reset_pushed()

            if is_coro:
                cohdl.coroutine_step(coro)
            elif step_cond():
                fn()

        wrapper.__name__ = trigger.__name__

        cohdl.sequential_context(
            _prefix_wrapper(parent_prefix, wrapper),
            name=trigger.__name__,
            attributes=attributes,
            source_location=SourceLocation.from_function(wrapped_fn),
            captured_functions=None if capture_lazy else [callable],
        )

        return trigger

    def helper(fn):
        nonlocal wrapped_fn
        wrapped_fn = fn if wrapped_fn is None else wrapped_fn
        is_coro = inspect.iscoroutinefunction(fn)

        if is_coro:
            callable = coro = fn()
        else:
            coro = None
            callable = fn

        if reset is None:

            def wrapper():
                cohdl.sensitivity.list(trigger.signal())
                if trigger:
                    if step_cond():
                        cohdl.reset_pushed()

                        if is_coro:
                            cohdl.coroutine_step(coro)
                        else:
                            fn()

            wrapper.__name__ = fn.__name__

            cohdl.sequential_context(
                _prefix_wrapper(parent_prefix, wrapper),
                name=fn.__name__,
                attributes=attributes,
                source_location=SourceLocation.from_function(wrapped_fn),
                captured_functions=None if capture_lazy else [callable],
            )

        elif reset.is_async():

            def wrapper():
                cohdl.sensitivity.list(trigger.signal(), reset.signal())
                if reset:
                    cohdl.reset_context()
                    for reset_fn in on_reset:
                        reset_fn()
                elif trigger:
                    if step_cond():
                        cohdl.reset_pushed()

                        if is_coro:
                            cohdl.coroutine_step(coro)
                        else:
                            fn()

            wrapper.__name__ = fn.__name__

            cohdl.sequential_context(
                _prefix_wrapper(parent_prefix, wrapper),
                name=fn.__name__,
                attributes=attributes,
                source_location=SourceLocation.from_function(wrapped_fn),
                captured_functions=None if capture_lazy else [callable],
            )

        else:

            def wrapper():
                cohdl.sensitivity.list(trigger.signal())
                if trigger:
                    if reset:
                        cohdl.reset_context()
                        for reset_fn in on_reset:
                            reset_fn()
                    elif step_cond():
                        cohdl.reset_pushed()
                        if is_coro:
                            cohdl.coroutine_step(coro)
                        else:
                            fn()

            wrapper.__name__ = fn.__name__

            cohdl.sequential_context(
                _prefix_wrapper(parent_prefix, wrapper),
                name=fn.__name__,
                attributes=attributes,
                source_location=SourceLocation.from_function(wrapped_fn),
                captured_functions=None if capture_lazy else [callable],
            )

    return helper


def sequential(
    trigger=None,
    /,
    reset=None,
    *,
    step_cond=None,
    comment=None,
    on_reset=None,
    attributes: dict | None = None,
    capture_lazy: bool = False,
):
    if isinstance(trigger, Clock):
        return SequentialContext(
            clk=trigger,
            reset=reset,
            step_cond=step_cond,
            comment=comment,
            on_reset=on_reset,
            attributes=attributes,
            capture_lazy=capture_lazy,
        )

    return _sequential_impl(
        trigger,
        reset=reset,
        step_cond=step_cond,
        comment=comment,
        on_reset=on_reset,
        attributes=attributes,
        capture_lazy=capture_lazy,
        wrapped_fn=None,
    )


def concurrent_assign(target, source):
    @concurrent()
    def logic():
        nonlocal target
        target <<= source


def concurrent_eval(target, fn, *args, **kwargs):
    @concurrent
    def logic():
        nonlocal target
        target <<= fn(*args, **kwargs)


def concurrent_call(fn, *args, **kwargs):
    @concurrent
    def logic():
        fn(*args, **kwargs)


class _ContextData:
    def __init__(
        self,
        ctx: SequentialContext,
        executors_before: list[Executor],
        executors_after: list[Executor],
    ):
        self.ctx = ctx
        self.executors_before = executors_before
        self.executors_after = executors_after


_current_context: SequentialContext | None = None
_current_context_data: _ContextData | None = None


class SequentialContext:
    @staticmethod
    @pyeval
    def current() -> SequentialContext | None:
        return _current_context

    @staticmethod
    @pyeval
    def current_data():
        return _current_context_data

    @staticmethod
    @pyeval
    def _enter_context(ctx: SequentialContext, data: _ContextData | None = None):
        global _current_context, _current_context_data
        _current_context = ctx
        _current_context_data = data

    @staticmethod
    @pyeval
    def _exit_context():
        global _current_context, _current_context_data
        _current_context = None
        _current_context_data = None

    def __init__(
        self,
        clk: Clock,
        reset: Reset | None = None,
        *,
        step_cond=None,
        comment=None,
        on_reset=None,
        attributes: dict | None = None,
        capture_lazy: bool = False,
    ):
        assert isinstance(clk, Clock)
        assert reset is None or isinstance(reset, Reset)

        self._clk = clk
        self._reset = reset
        self._step_cond = step_cond
        self._comment = comment
        self._on_reset = on_reset
        self._attributes = attributes
        self._capture_lazy = capture_lazy

    def copy(self):
        return type(self)(
            clk=self._clk,
            reset=self._reset,
            step_cond=self._step_cond,
            comment=self._comment,
            on_reset=self._on_reset,
            attributes=self._attributes,
            capture_lazy=self._capture_lazy,
        )

    def clk(self):
        return self._clk

    def reset(self):
        return self._reset

    def step_cond(self):
        return self._step_cond

    def with_params(
        self,
        *,
        clk: Clock | None = None,
        reset: Reset | None = None,
        step_cond=None,
        on_reset=None,
    ):
        return SequentialContext(
            clk=self._clk if clk is None else clk,
            reset=self._reset if reset is None else reset,
            step_cond=self._step_cond if step_cond is None else step_cond,
            on_reset=self._on_reset if on_reset is None else on_reset,
            attributes=self._attributes,
        )

    def or_reset(
        self,
        cond: Bit | None = None,
        *,
        expr=None,
        active_low: bool = False,
        is_async: bool | None = None,
    ):
        combined_reset = Signal[Bit]()

        if cond is not None:
            assert (
                expr is None
            ), "the arguments `cond` and `expr` are mutually exclusive"
            expr = lambda: cond
        else:
            assert (
                expr is not None
            ), "the arguments `cond` and `expr` are mutually exclusive"

        if self._reset is None:

            @concurrent
            def logic():
                nonlocal combined_reset
                combined_reset <<= expr()

        else:
            is_async = is_async if is_async is not None else self._reset.is_async()

            @concurrent
            def logic():
                nonlocal combined_reset

                if active_low:
                    # combined is in reset state (low) when one of the inputs is in reset state
                    combined_reset <<= self._reset.active_low_signal() and expr()
                else:
                    # combined is in reset state (high) when one of the inputs is in reset state
                    combined_reset <<= self._reset.active_high_signal() or expr()

        return SequentialContext(
            self._clk,
            Reset(combined_reset, active_low=active_low, is_async=is_async),
            attributes=self._attributes,
        )

    def and_reset(
        self,
        cond: Bit | None = None,
        *,
        expr=None,
        active_low: bool = False,
        is_async: bool | None = None,
    ):
        combined_reset = Signal[Bit]()

        if cond is not None:
            assert (
                expr is None
            ), "the arguments `cond` and `expr` are mutually exclusive"
            expr = lambda: cond
        else:
            assert (
                expr is not None
            ), "the arguments `cond` and `expr` are mutually exclusive"

        if self._reset is None:

            @concurrent
            def logic():
                nonlocal combined_reset
                combined_reset <<= expr()

        else:
            is_async = is_async if is_async is not None else self._reset.is_async()

            @concurrent
            def logic():
                nonlocal combined_reset

                if active_low:
                    combined_reset <<= self._reset.active_low_signal() or expr()
                else:
                    combined_reset <<= self._reset.active_high_signal() and expr()

        return SequentialContext(
            self._clk,
            Reset(combined_reset, active_low=active_low, is_async=is_async),
            attributes=self._attributes,
        )

    def __call__(
        self,
        fn=None,
        *,
        on_reset=None,
        executors: list[Executor] | None = None,
        attributes: dict | None = None,
    ):
        # copy the sequential context so each invocation has a distinct,
        # associated std.SequentialContext
        cpy = self.copy()

        if attributes is None:
            attributes = cpy._attributes
        elif cpy._attributes is not None:
            attributes = {**cpy._attributes, **attributes}
        else:
            pass

        executors = [] if executors is None else executors

        data = _ContextData(
            cpy,
            executors_before=[
                e for e in executors if e._mode is ExecutorMode.immediate_before
            ],
            executors_after=[
                e for e in executors if e._mode is ExecutorMode.immediate_after
            ],
        )

        def convert_executors(executors: list[Executor], mode: ExecutorMode):
            for executor in executors:
                assert executor._mode is mode, "internal error: executor mode mismatch"

                cohdl.coroutine_step(executor.executor_statemachine())

                if not executor._independent:
                    executor._start @= False

        def helper(fn):
            if inspect.iscoroutinefunction(fn):

                def context_fn():
                    cpy._enter_context(cpy, data)
                    convert_executors(
                        data.executors_before, mode=ExecutorMode.immediate_before
                    )
                    cohdl.coroutine_step(fn())
                    convert_executors(
                        data.executors_after, mode=ExecutorMode.immediate_after
                    )
                    cpy._exit_context()

            else:

                def context_fn():
                    cpy._enter_context(cpy, data)
                    convert_executors(
                        data.executors_before, mode=ExecutorMode.immediate_before
                    )
                    fn()
                    convert_executors(
                        data.executors_after, mode=ExecutorMode.immediate_after
                    )
                    cpy._exit_context()

            context_fn.__name__ = fn.__name__
            return _sequential_impl(
                cpy._clk,
                cpy._reset,
                step_cond=cpy._step_cond,
                on_reset=on_reset,
                comment=cpy._comment,
                attributes=attributes,
                capture_lazy=cpy._capture_lazy,
                wrapped_fn=fn,
            )(context_fn)

        if fn is not None:
            return helper(fn)
        else:
            return helper


Context = SequentialContext

#
#
#
#
#
#
#


class ExecutorMode(enum.Enum):
    parallel_process = enum.auto()
    immediate_after = enum.auto()
    immediate_before = enum.auto()


class Executor:
    """
    Executor wraps a coroutine in a _sequential_impl process
    and provides methods to start the execution from a different
    context.
    """

    @pyeval
    def _check_context(self):
        if not self._mode is ExecutorMode.parallel_process:
            current_ctx = SequentialContext.current()
            current_data = SequentialContext.current_data()
            assert (
                current_ctx is not None
            ), f"Executor with mode {self._mode} can only be used in functions marked with std.SequentialContext"

            if self._mode is ExecutorMode.immediate_before:
                assert (
                    self in current_data.executors_before
                ), "all used executors with mode {ExecutorMode.immediate_before} must be specified in the std.SequentialContext decorator of the enclosing function"
            else:
                assert (
                    self._mode is ExecutorMode.immediate_after
                ), f"invalid executor mode {self._mode}"
                if self not in current_data.executors_after:
                    current_data.executors_after.append(self)

    async def executor_statemachine(self):
        if not self._independent:
            while not self._start:
                self._process_ready._assign_(True, AssignMode.AUTO)
            self._process_ready._assign_(False, AssignMode.AUTO)

            if self._result is None:
                await self._action(*self._args, **self._kwargs)
            else:
                self._result._assign_(
                    await self._action(*self._args, **self._kwargs), AssignMode.AUTO
                )
            self._process_ready._assign_(True, AssignMode.AUTO)
        else:
            await self._action(*self._args, **self._kwargs)

    @pyeval
    def __init__(
        self,
        ctx: SequentialContext | None,
        mode: ExecutorMode,
        action,
        result=None,
        args: list | None = None,
        kwargs: dict | None = None,
        independent: bool = False,
    ):
        self._mode = mode
        self._ctx = ctx
        self._action = action

        args = args if args is not None else []
        kwargs = kwargs if kwargs is not None else {}

        self._args = args
        self._kwargs = kwargs
        self._result = result
        self._independent = independent

        if ctx is None:
            if not independent:
                self._start = Variable[Bit](False, name="executor_start")
                self._process_ready = Variable[Bit](
                    False, name="executor_process_ready"
                )
        else:
            assert (
                mode is ExecutorMode.parallel_process
            ), "mode must be `parallel_process` because `ctx` was specified"

            if not independent:
                self._start = Signal[Bit](False, name="executor_start")
                self._process_ready = Signal[Bit](False, name="executor_process_ready")

            @ctx
            async def proc():
                await self.executor_statemachine()

    @classmethod
    @pyeval
    def make_parallel(cls, ctx: SequentialContext, action, result, *args, **kwargs):
        return cls(
            ctx=ctx,
            mode=ExecutorMode.parallel_process,
            action=action,
            result=result,
            args=args,
            kwargs=kwargs,
        )

    @classmethod
    def make_before(cls, action, result, *args, **kwargs):
        return cls(
            ctx=None,
            mode=ExecutorMode.immediate_before,
            action=action,
            result=result,
            args=args,
            kwargs=kwargs,
        )

    @classmethod
    @pyeval
    def make_after(cls, action, result, *args, **kwargs):
        return cls(
            ctx=None,
            mode=ExecutorMode.immediate_after,
            action=action,
            result=result,
            args=args,
            kwargs=kwargs,
        )

    @classmethod
    @pyeval
    def make_independent_before(cls, action, result, *args, **kwargs):
        return cls(
            ctx=None,
            mode=ExecutorMode.immediate_before,
            action=action,
            result=result,
            args=args,
            kwargs=kwargs,
            independent=True,
        )

    @classmethod
    @pyeval
    def make_independent_after(cls, action, result, *args, **kwargs):
        return cls(
            ctx=None,
            mode=ExecutorMode.immediate_after,
            action=action,
            result=result,
            args=args,
            kwargs=kwargs,
            independent=True,
        )

    def start(self, *args, **kwargs):
        self._check_context()
        assert self.ready()
        assert len(self._args) == len(args)

        for target, src in zip(self._args, args):
            target._assign_(src, AssignMode.AUTO)

        for name, val in kwargs.items():
            self._kwargs[name]._assign_(val, AssignMode.AUTO)

        if self._mode is ExecutorMode.parallel_process:
            self._start ^= True
        else:
            self._start @= True

    async def exec(self, *args, **kwargs):
        self._check_context()
        assert self.ready()
        self.start(*args, **kwargs)
        await self.ready()

        if self._result is not None:
            return self._result.copy()
        else:
            return None

    @cohdl.expr_fn
    def ready(self):
        return self._process_ready and not self._start

    def result(self):
        assert self.ready()
        return self._result.copy()

    def __hash__(self):
        return id(self)


def at_end_of_context(fn):
    Executor.make_independent_after(fn, None)._check_context()



from __future__ import annotations


class TextBlock:
    def __init__(
        self,
        content: str | TextBlock | list[str | TextBlock],
        *,
        title=None,
        indent=False,
        trailer=None,
    ):
        self._title = title
        self._indent = indent if title is None else True
        self._trailer = trailer

        self._content = []
        self.add(content)

    def add(self, elem: str | TextBlock | list[str | TextBlock]):
        if isinstance(elem, list):
            for i in elem:
                self.add(i)
        else:
            assert isinstance(elem, (str, TextBlock))
            self._content.append(elem)

    def _dump_impl(self, indent, step=2):
        result = []

        initial_indent = indent

        if self._title is not None:
            result.append(indent + self._title)

        if self._indent:
            indent = indent + " " * step

        for elem in self._content:
            if isinstance(elem, str):
                result.append(indent + elem)
            else:
                result.extend(elem._dump_impl(indent, step))

        if self._trailer is not None:
            result.append(initial_indent + self._trailer)

        return result

    def dump(self, indent="", step=2):
        return "\n".join(self._dump_impl(indent, step))

    def __str__(self):
        return self.dump()

    def __repr__(self) -> str:
        return self.dump()


class IndentBlock(TextBlock):
    def __init__(self, content: str | TextBlock | list[str | TextBlock], *, title=None):
        super().__init__(content, title=title, indent=True)



from __future__ import annotations

import inspect

from cohdl._core import AssignMode, Null, Full
from cohdl._core._intrinsic import _intrinsic
from cohdl.utility import TextBlock

from ._core_utility import (
    count_bits,
    from_bits,
    to_bits,
    concat,
    Value,
    Ref,
)

from ._assignable_type import AssignableType
from ._template import Template, _TemplateMode
from ._prefix import NamedQualifier


@_intrinsic
def _is_in(value, container):
    return value in container


@_intrinsic
def _args_to_kwargs(kwargs, args, names):
    for arg, name in zip(args, names):
        assert (
            name not in kwargs
        ), f"duplicate definition of '{name}' as positional and keyword argument"
        kwargs[name] = arg


@_intrinsic
def _make_serializable(cls: Record):
    if hasattr(cls, "_cohdlstd_bitcount"):
        return

    slice_map = {}
    elem_start = 0

    for name, elem_type in cls._cohdlstd_record_annotations.items():
        width = count_bits(elem_type)
        slice_map[name] = slice(elem_start + width - 1, elem_start)
        elem_start = elem_start + width

    assert elem_start != 0, "record does not contain any elements to serialize"

    cls._cohdlstd_bitcount = elem_start
    cls._cohdlstd_slice_map = slice_map


@_intrinsic
def _get_reverse_elem_list(self: Record):
    # returns the elements of this record in reverse
    # declaration order so they can be serialized using std.concat

    return [getattr(self, name) for name in self._cohdlstd_record_annotations.keys()][
        ::-1
    ]


class Record(AssignableType, Template):
    # _cohdlstd_bitcount: int
    # _cohdlstd_slice_map: dict[str, slice]

    @classmethod
    def _make_ref_(cls, *args, **kwargs):
        return cls(*args, **kwargs, _qualifier_=Ref)

    @_intrinsic
    def __new__(cls, *args, **kwargs):
        # override __new__ because Template.__new__ does
        # not work for non-template Records
        inst = object.__new__(cls)
        return inst

    def _assign_(self, source, mode: AssignMode) -> None:
        if isinstance(source, dict):
            for name, value in source.items():
                getattr(self, name)._assign_(value, mode)
        elif source is Null or source is Full:
            for name, value in self.__dict__.items():
                value._assign_(source, mode)
        else:
            assert isinstance(source, type(self))
            self._assign_(source.__dict__, mode)

    def __init_subclass__(cls) -> None:
        if cls._template_meta_.mode is _TemplateMode.ROOT:
            # record derived without template arguments
            module_dict = inspect.getmodule(cls).__dict__

            annotations = {
                name: eval(value, module_dict)
                for name, value in cls.__annotations__.items()
            }
        else:
            # record derived with template arguments

            annotations = cls._template_meta_.annotations
            annotations = {} if annotations is None else annotations

        if hasattr(cls, "_cohdlstd_record_annotations"):
            overlap = cls._cohdlstd_record_annotations.keys() & annotations.keys()
            assert (
                len(overlap) == 0
            ), f"record elements '{overlap}' would be overwritten"
            annotations = {**cls._cohdlstd_record_annotations, **annotations}

        cls._cohdlstd_record_annotations = annotations

    def __init__(self, *args, _qualifier_=Value, **kwargs):
        elem_types = self._cohdlstd_record_annotations

        if len(args) == len(kwargs) == 0:
            # default constructor
            for name, elem_type in elem_types.items():
                setattr(self, name, NamedQualifier[_qualifier_, name][elem_type]())
        elif len(args) == 1 and len(kwargs) == 0:
            arg = args[0]

            if arg is Null or arg is Full:
                for name, elem_type in elem_types.items():
                    setattr(
                        self, name, NamedQualifier[_qualifier_, name][elem_type](arg)
                    )
            else:
                # copy constructor
                assert type(arg) is type(self), "type mismatch {} != {}".format(
                    type(arg), type(self)
                )
                self.__init__(**arg.__dict__, _qualifier_=_qualifier_)
        else:
            _args_to_kwargs(kwargs, args, elem_types)

            if len(kwargs) != len(elem_types):
                missing = elem_types.keys() - kwargs.keys()
                additional = kwargs.keys() - elem_types.keys()

                if len(missing) == 0:
                    raise AssertionError(
                        f"argument error: additional args = {additional}"
                    )

                if len(additional) == 0:
                    raise AssertionError(f"argument error: missing args = {missing}")

                raise AssertionError(
                    f"argument error: missing args = {missing}, additional args = {additional}"
                )

            for name, arg in kwargs.items():
                expected_type = elem_types[name]
                setattr(self, name, _qualifier_[expected_type](arg))

    @classmethod
    @_intrinsic
    def _count_bits_(cls):
        _make_serializable(cls)
        return cls._cohdlstd_bitcount

    @classmethod
    def _from_bits_(cls, bits, qualifier):
        _make_serializable(cls)
        assert bits.width == cls._count_bits_()

        result = cls(
            **{
                name: from_bits[elem_type](
                    bits[cls._cohdlstd_slice_map[name]], qualifier
                )
                for name, elem_type in cls._cohdlstd_record_annotations.items()
            },
            # use Ref qualifier because outer qualifier has already been
            # applied during dict construction
            _qualifier_=Ref,
        )

        return result

    def _to_bits_(self):
        _make_serializable(type(self))
        return concat(*[to_bits(elem) for elem in _get_reverse_elem_list(self)])

    @_intrinsic
    def __repr__(self):
        args = ", ".join(f"{name}={value!r}" for name, value in self.__dict__.items())
        return f"{type(self).__name__}({args})"

    @_intrinsic
    def _str_impl(self, elem_name=None):
        type_name = type(self).__name__

        return TextBlock(
            title=type_name if elem_name is None else f"{elem_name}={type_name}",
            content=[
                (
                    f"{name}={value!s}"
                    if not isinstance(value, Record)
                    else value._str_impl(name)
                )
                for name, value in self.__dict__.items()
            ],
            indent=True,
        )

    @_intrinsic
    def __str__(self):
        return self._str_impl().dump()

    def __eq__(self, other):
        return all(
            [value == other.__dict__[name] for name, value in self.__dict__.items()]
        )

    def __ne__(self, other):
        return any(
            [value != other.__dict__[name] for name, value in self.__dict__.items()]
        )



from __future__ import annotations

import typing
from typing import overload

import enum

from cohdl._core._bit import Bit
from cohdl._core._bit_vector import BitVector
from cohdl._core._unsigned import Unsigned
from cohdl._core._signed import Signed
from cohdl._core._intrinsic_operations import AssignMode
from cohdl._core._integer import Integer

T = typing.TypeVar("T", Bit, BitVector)
U = typing.TypeVar("U", Bit, BitVector)
VEC_T = typing.TypeVar("VEC_T", BitVector)
DIR = typing.TypeVar("DIR")
Self = typing.TypeVar("Self")

class Attribute:
    name: str
    attr_type: type

    def __init__(self, value):
        self.value = value

class TypeQualifierBase(typing.Generic[T]):
    def decay(val): ...

class TypeQualifier(typing.Generic[T], TypeQualifierBase[T]):
    @property
    def type(cls): ...
    @property
    def qualifier(self): ...
    def decay(val): ...
    @property
    def width(self) -> int: ...

    #
    #
    #

    def __init__(
        self,
        value=None,
        *,
        name: str | None = None,
        attributes: dict | None = None,
        noreset: bool = False,
        maybe_uninitialized: bool = False,
    ) -> None:
        """
        Create a new type qualified object with an optional default `value`.

        `name` is only a hint. The compiler will add a counter suffix to resolve name collisions.
        `attributes` are currently unused.

        Objects declared with `noreset` will not be automatically reset in
        synthesizable contexts (the default value is still used during initialization).

        CoHDL perform some basic checks to ensure, that all locally declared objects
        are initialized before they are used. Set `maybe_uninitialized` to True to
        opt out of these checks.
        """

    def name(self) -> str: ...
    def has_default(self) -> bool: ...
    def default(self): ...
    def __len__(self) -> int: ...
    def __iter__(self): ...
    def get(self) -> T: ...
    @overload
    def __getitem__(self, arg: int) -> TypeQualifier[Bit]: ...
    @overload
    def __getitem__(self, arg: slice) -> TypeQualifier[BitVector]: ...
    @overload
    def __getitem__(self, arg: tuple | list) -> Temporary[BitVector]: ...
    @overload
    def lsb(self, count: None = None, rest: None = None) -> TypeQualifier[Bit]: ...
    @overload
    def lsb(
        self, count: None | int = None, rest: None | int = None
    ) -> TypeQualifier[BitVector]: ...
    @overload
    def msb(self, count: None = None, rest: None = None) -> TypeQualifier[Bit]: ...
    @overload
    def msb(
        self, count: None | int = None, rest: None | int = None
    ) -> TypeQualifier[BitVector]: ...
    @overload
    def left(self, count: None = None, rest: None = None) -> TypeQualifier[Bit]: ...
    @overload
    def left(
        self, count: None | int = None, rest: None | int = None
    ) -> TypeQualifier[BitVector]: ...
    @overload
    def right(self, count: None = None, rest: None = None) -> TypeQualifier[Bit]: ...
    @overload
    def right(
        self, count: None | int = None, rest: None | int = None
    ) -> TypeQualifier[BitVector]: ...

    #
    #
    #

    def __bool__(self) -> bool: ...
    def __and__(self, other) -> Temporary[T]: ...
    def __rand__(self, other) -> Temporary[T]: ...
    def __or__(self, other) -> Temporary[T]: ...
    def __ror__(self, other) -> Temporary[T]: ...
    def __xor__(self, other) -> Temporary[T]: ...
    def __rxor__(self, other) -> Temporary[T]: ...
    def __matmul__(self, other) -> Temporary[T]: ...
    def __rmatmul__(self, other) -> Temporary[T]: ...
    #
    # numeric operators
    #

    def __add__(self, other) -> Temporary[T]: ...
    def __radd__(self, other) -> Temporary[T]: ...
    def __sub__(self, other) -> Temporary[T]: ...
    def __rsub__(self, other) -> Temporary[T]: ...
    def __mul__(self, other) -> Temporary[T]: ...
    def __rmul__(self, other) -> Temporary[T]: ...
    def __floordiv__(self, other) -> Temporary[T]: ...
    def __rfloordiv__(self, other) -> Temporary[T]: ...
    def __mod__(self, other) -> Temporary[T]: ...
    def __rmod__(self, other) -> Temporary[T]: ...

    #
    # shift operators
    #

    def __lshift__(self, rhs: Unsigned | int | Integer) -> Temporary[T]: ...
    def __rshift__(self, rhs: Unsigned | int | Integer) -> Temporary[T]: ...

    #
    # compare
    #

    def __eq__(self, other) -> bool: ...
    def __ne__(self, other) -> bool: ...
    def __lt__(self, other) -> bool: ...
    def __gt__(self, other) -> bool: ...
    def __le__(self, other) -> bool: ...
    def __ge__(self, other) -> bool: ...
    #
    # unary operators
    #

    def __inv__(self: Self) -> Temporary[T]: ...
    def __neg__(self: Self) -> Temporary[T]: ...
    def __pos__(self: Self) -> Temporary[T]: ...
    #
    # casts
    #
    @property
    def unsigned(self) -> TypeQualifier[Unsigned]: ...
    @unsigned.setter
    def unsigned(self, value: TypeQualifier[Unsigned] | Unsigned | int): ...
    @property
    def signed(self) -> TypeQualifier[Signed]: ...
    @signed.setter
    def signed(self, value: TypeQualifier[Signed] | Signed | int): ...
    @property
    def bitvector(self) -> TypeQualifier[BitVector]: ...
    @bitvector.setter
    def bitvector(self, value: TypeQualifier[BitVector] | BitVector | str): ...

    #
    # helper methods
    #

    def resize(
        self, target_width: int | None = None, *, zeros: int = 0
    ) -> Temporary[T]: ...
    def copy(self) -> Temporary[T]: ...

class Signal(typing.Generic[T], TypeQualifier[T]):
    def __init__(
        self,
        value=None,
        *,
        name: str | None = None,
        attributes: dict | None = None,
        delayed_init: bool = False,
        maybe_uninitialized: bool = False,
    ) -> None:
        """
        Create a new Signal with an optional default `value`.

        `name` is only a hint. The compiler will add a counter suffix to resolve name collisions.
        `attributes` are currently unused.

        `delayed_init` only has an effect for signals created locally in a sequential context.
        By default these initializations are *NOT* the same as a signal assignment.
        Instead initialization takes place immediately like variable assignment.
        When `delayed_init` is set to true initialization behaves like a signal assignment
        and takes one clock cycle.

        CoHDL perform some basic checks to ensure, that all locally declared objects
        are initialized before they are used. Set `maybe_uninitialized` to True to
        opt out of these checks.

        >>> @std.sequential(clk)
        >>> def example():
        >>>     a = Signal[Bit](input)
        >>>     b = Signal[Bit](a)
        >>>     c = Signal[Bit](b)
        >>>
        >>>     # this assertion always holds
        >>>     assert input == a == b == c
        >>>
        >>>     x = Signal[Bit](input, delayed_init=True)
        >>>     y = Signal[Bit](x, delayed_init=True)
        >>>
        >>>     assert x == prev(input)
        >>>     assert x == prev(prev(input))
        """
    #
    #
    #
    @overload
    def __getitem__(self, arg: int) -> Signal[Bit]: ...
    @overload
    def __getitem__(self, arg: slice) -> Signal[BitVector]: ...
    @overload
    def __getitem__(self, arg: tuple | list) -> Temporary[BitVector]: ...
    @overload
    def lsb(self, count: None = None, rest: None = None) -> Signal[Bit]: ...
    @overload
    def lsb(
        self, count: None | int = None, rest: None | int = None
    ) -> Signal[BitVector]: ...
    @overload
    def msb(self, count: None = None, rest: None = None) -> Signal[Bit]: ...
    @overload
    def msb(
        self, count: None | int = None, rest: None | int = None
    ) -> Signal[BitVector]: ...
    @overload
    def right(self, count: None = None, rest: None = None) -> Signal[Bit]: ...
    @overload
    def right(
        self, count: None | int = None, rest: None | int = None
    ) -> Signal[BitVector]: ...
    @overload
    def left(self, count: None = None, rest: None = None) -> Signal[Bit]: ...
    @overload
    def left(
        self, count: None | int = None, rest: None | int = None
    ) -> Signal[BitVector]: ...

    #
    # casts
    #
    @property
    def unsigned(self: Signal[VEC_T]) -> Signal[Unsigned]: ...
    @unsigned.setter
    def unsigned(
        self: Signal[VEC_T], value: TypeQualifier[Unsigned] | Unsigned | int
    ): ...
    @property
    def signed(self: Signal[VEC_T]) -> Signal[Signed]: ...
    @signed.setter
    def signed(self, value: TypeQualifier[Signed] | Signed | int): ...
    @property
    def bitvector(self: Signal[VEC_T]) -> Signal[BitVector]: ...
    @bitvector.setter
    def bitvector(
        self: Signal[VEC_T], value: TypeQualifier[BitVector] | BitVector | str
    ): ...

    #
    #
    #

    @property
    def next(self) -> typing.NoReturn: ...
    @next.setter
    def next(self, value) -> None: ...
    def __ilshift__(self: Self, value) -> Self: ...
    @property
    def push(self) -> typing.NoReturn: ...
    @push.setter
    def push(self, value) -> None: ...
    def __ixor__(self: Self, value) -> Self: ...
    #
    #
    def _assign_(self, value, assign_mode: AssignMode) -> None: ...

class Port(typing.Generic[T, DIR], Signal[T]):
    #
    #
    #
    @overload
    def __getitem__(self, arg: int) -> Port[Bit, DIR]: ...
    @overload
    def __getitem__(self, arg: slice) -> Port[BitVector, DIR]: ...
    @overload
    def __getitem__(self, arg: tuple | list) -> Temporary[BitVector]: ...
    @overload
    def lsb(self, count: None = None, rest: None = None) -> Port[Bit, DIR]: ...
    @overload
    def lsb(
        self, count: None | int = None, rest: None | int = None
    ) -> Port[BitVector, DIR]: ...
    @overload
    def msb(self, count: None = None, rest: None = None) -> Port[Bit, DIR]: ...
    @overload
    def msb(
        self, count: None | int = None, rest: None | int = None
    ) -> Port[BitVector, DIR]: ...
    @overload
    def right(self, count: None = None, rest: None = None) -> Port[Bit, DIR]: ...
    @overload
    def right(
        self, count: None | int = None, rest: None | int = None
    ) -> Port[BitVector, DIR]: ...
    @overload
    def left(self, count: None = None, rest: None = None) -> Port[Bit, DIR]: ...
    @overload
    def left(
        self, count: None | int = None, rest: None | int = None
    ) -> Port[BitVector, DIR]: ...

    #
    # casts
    #
    @property
    def unsigned(self: Port[VEC_T, DIR]) -> Port[Unsigned, DIR]: ...
    @unsigned.setter
    def unsigned(
        self: Port[VEC_T, DIR], value: TypeQualifier[Unsigned] | Unsigned | int
    ): ...
    @property
    def signed(self: Port[VEC_T, DIR]) -> Port[Signed, DIR]: ...
    @signed.setter
    def signed(self: Port[VEC_T, DIR], value: TypeQualifier[Signed] | Signed | int): ...
    @property
    def bitvector(self: Port[VEC_T, DIR]) -> Port[BitVector, DIR]: ...
    @bitvector.setter
    def bitvector(
        self: Port[VEC_T, DIR], value: TypeQualifier[BitVector] | BitVector | str
    ): ...
    #
    #
    #

    class Direction(enum.Enum):
        class INPUT: ...
        class OUTPUT: ...
        class INOUT: ...

        def is_input(self) -> bool: ...
        def is_output(self) -> bool: ...
        def is_inout(self) -> bool: ...

    @classmethod
    def is_input(cls) -> bool: ...
    @classmethod
    def is_output(cls) -> bool: ...
    @classmethod
    def is_inout(cls) -> bool: ...
    @staticmethod
    def input(
        Wrapped: type[U], *, name: str | None = None
    ) -> Port[U, Port.Direction.INPUT]: ...
    @staticmethod
    def output(
        Wrapped: type[U], *, default=None, name: str | None = None
    ) -> Port[U, Port.Direction.OUTPUT]: ...
    @staticmethod
    def inout(
        Wrapped: type[U], *, default=None, name: str | None = None
    ) -> Port[U, Port.Direction.INOUT]: ...
    @classmethod
    def direction(cls) -> Port.Direction: ...

class Variable(typing.Generic[T], TypeQualifier[T]):
    #
    #
    #
    @overload
    def __getitem__(self, arg: int) -> Variable[Bit]: ...
    @overload
    def __getitem__(self, arg: slice) -> Variable[BitVector]: ...
    @overload
    def __getitem__(self, arg: tuple | list) -> Temporary[BitVector]: ...
    @overload
    def lsb(self, count: None = None, rest: None = None) -> Variable[Bit]: ...
    @overload
    def lsb(
        self, count: None | int = None, rest: None | int = None
    ) -> Variable[BitVector]: ...
    @overload
    def msb(self, count: None = None, rest: None = None) -> Variable[Bit]: ...
    @overload
    def msb(
        self, count: None | int = None, rest: None | int = None
    ) -> Variable[BitVector]: ...
    @overload
    def right(self, count: None = None, rest: None = None) -> Variable[Bit]: ...
    @overload
    def right(
        self, count: None | int = None, rest: None | int = None
    ) -> Variable[BitVector]: ...
    @overload
    def left(self, count: None = None, rest: None = None) -> Variable[Bit]: ...
    @overload
    def left(
        self, count: None | int = None, rest: None | int = None
    ) -> Variable[BitVector]: ...

    #
    # casts
    #
    @property
    def unsigned(self) -> Variable[Unsigned]: ...
    @unsigned.setter
    def unsigned(self, value: TypeQualifier[Unsigned] | Unsigned | int): ...
    @property
    def signed(self) -> Variable[Signed]: ...
    @signed.setter
    def signed(self, value: TypeQualifier[Signed] | Signed | int): ...
    @property
    def bitvector(self) -> Variable[BitVector]: ...
    @bitvector.setter
    def bitvector(self, value: TypeQualifier[BitVector] | BitVector | str): ...
    #
    #
    #
    @property
    def value(self): ...
    @value.setter
    def value(self, value) -> typing.NoReturn: ...
    def __imatmul__(self: Self, value) -> Self: ...

    #
    #
    def _assign_(self, value, assign_mode: AssignMode) -> None: ...

class Temporary(typing.Generic[T], TypeQualifier[T]):
    #
    #
    #
    @overload
    def __getitem__(self, arg: int) -> Temporary[Bit]: ...
    @overload
    def __getitem__(self, arg: slice) -> Temporary[BitVector]: ...
    @overload
    def __getitem__(self, arg: tuple | list) -> Temporary[BitVector]: ...
    @overload
    def lsb(self, count: None = None, rest: None = None) -> Temporary[Bit]: ...
    @overload
    def lsb(
        self, count: None | int = None, rest: None | int = None
    ) -> Temporary[BitVector]: ...
    @overload
    def msb(self, count: None = None, rest: None = None) -> Temporary[Bit]: ...
    @overload
    def msb(
        self, count: None | int = None, rest: None | int = None
    ) -> Temporary[BitVector]: ...
    @overload
    def right(self, count: None = None, rest: None = None) -> Temporary[Bit]: ...
    @overload
    def right(
        self, count: None | int = None, rest: None | int = None
    ) -> Temporary[BitVector]: ...
    @overload
    def left(self, count: None = None, rest: None = None) -> Temporary[Bit]: ...
    @overload
    def left(
        self, count: None | int = None, rest: None | int = None
    ) -> Temporary[BitVector]: ...

    #
    # casts
    #
    @property
    def unsigned(self) -> Temporary[Unsigned]: ...
    @unsigned.setter
    def unsigned(self, value: TypeQualifier[Unsigned] | Unsigned | int): ...
    @property
    def signed(self) -> Temporary[Signed]: ...
    @signed.setter
    def signed(self, value: TypeQualifier[Signed] | Signed | int): ...
    @property
    def bitvector(self) -> Temporary[BitVector]: ...
    @bitvector.setter
    def bitvector(self, value: TypeQualifier[BitVector] | BitVector | str): ...

class Generic: ...



from cohdl import std
from cohdl import BitVector, Unsigned, Signed, Bit, Null, Full, Signal


class MetaData:
    def _config_(self, *args, **kwargs):
        ...

    def __init__(self):
        ...


class MetaUser:
    def __init__(self):
        ...

    def _wrapped_init_(self, *args, **kwargs):
        self._real_init_(*args, **kwargs)

        for name, ann in type(self).__annotations__.items():
            if hasattr(self, name):
                assert isinstance(getattr(self, name), ann)
            else:
                assert issubclass(ann, MetaData)
                setattr(self, ann())

            getattr(self, name)._meta_name_ = name
            getattr(self, name)._config_(*ann._meta_args_)


class Register(MetaData, MetaUser):
    _global_offset_: int

    def __init__(self):
        self._offset_: int

        if not hasattr(self, "_comment_"):
            self._comment_ = ""

        type(self).__a

    def _config_(self, args):
        if not isinstance(args, tuple):
            args = (args,)

        assert len(args) == 1 or len(args) == 2
        assert isinstance(args[0], int)

        self._offset_ = args[0]
        self._comment_ = args[1] if len(args) == 2 else self._comment_


class Field(MetaData):
    ...


class RegisterDevice(MetaData, MetaUser):
    _global_offset_: int
    ...


class Adder(RegisterDevice):
    """
    Example register with two input registers, that
    are added together and returned in a result register
    """

    inp_a: Register.W[0x00, "first summand"]
    inp_b: Register.W[0x04, "second summand"]
    result: Register.R[0x08, "result of inp_a+inp_b"]

    def _impl_concurrent_(self):
        self.result <<= self.inp_a + self.inp_b


#
#
#
#
#


class WaitRegister(Register):
    async def on_write(self, data: Unsigned, mask=Full):
        await std.wait_for(data.unsigned)


class RgbLed(reg.Register):
    "This register device controls the rgb leds"

    r: reg.FieldU[7:0]
    g: reg.FieldU[15:8]
    b: reg.FieldU[23:16]

    async def _impl_sequential_(self):
        await std.wait_for(std.us(5))
        self.cnt <<= self.cnt + 1
        self.rgb.red <<= cnt < self.r.val
        self.rgb.green <<= cnt < self.g.val
        self.rgb.blue <<= cnt < self.b.val

    def __init__(self, rgb):
        self.rgb = rgb
        self.cnt = Signal[Unsigned[8]](Null)


class SpiCtrlRegister(reg.Register):
    "This register controls the spi communication"

    count: reg.Field[5:2, 0]["asdf"]
    active: reg.Field[8]["set to '1' while a spi transaction is in progress"]

    def __init__(self):
        self.transaction_active = std.SyncFlag()

    def _impl_concurrent_(self):
        self.active <<= self.transaction_active.is_set()

    def on_write(self, data, mask=Full):
        view = self._data_view_(data, mask)

        if view.active:
            assert not self.transaction_active.is_set()
            self.transaction_active.set()

        self.count <<= view.count


class SpiMaster(reg.RegisterDevice):
    ctrl: SpiCtrlRegister[0]
    send_low: reg.W[4]
    send_high: reg.W[8]
    receive_low: reg.R[12]
    receive_high: reg.R[16]

    def __init__(
        self,
        spi: std.spi.Spi,
        clk_period=std.Duration,
    ):
        super().__init__()
        self.spi = spi
        self.clk_period = clk_period

    def _impl_(self, ctx):
        self.master = std.spi.SpiMaster(ctx, self.spi, clk_period=self.clk_period)
        self.send_data = Signal[BitVector[64]](Null)
        self.receive_data = Signal[BitVector[64]](Null)

    async def _impl_sequential_(self):
        async with self.ctrl.transaction_active:
            self.receive_data <<= await self.master.parallel_transaction(
                self.send_data, self.ctrl.count
            )

    def _impl_concurrent_(self):
        self.receive_low <<= self.receive_data[31:0]
        self.receive_high <<= self.receive_data[63:32]
        self.send_data[31:0] <<= self.send_low
        self.send_data[63:32] <<= self.send_high


class Wait(reg.Register):
    async def on_write(self, value: BitVector, mask=Full):
        std.wait_for(value.unsigned)


class Interface(reg.RegisterDevice):
    """
    Collection of all hardware interfaces.
    """

    rgb_a: RgbLed[0x100]
    rgb_b: RgbLed[0x104]

    spi: SpiMaster[0x200]

    def __init__(self, rgb_a, rgb_b, spi, spi_period):
        self.rgb_a = RgbLed(rgb_a)
        self.rgb_b = RgbLed(rgb_b)
        self.spi = SpiMaster(spi, spi_period)


class Args:
    def __init__(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs


def _main_():
    con = interconnect.reserve(0x10000000, 10)

    con.registers()

    all_regs(
        {
            0: reg.R(switches, "switches"),
            4: reg.W(leds, "leds"),
            8: Wait(),
            12: reg.R(buttons, "buttons"),
            0x100: ethernet.EthernetDevice(
                ctx=eth_ctx,
                crsdv=eth.crsdv,
                rxd0=eth.rxd0,
                rxd1=eth.rxd1,
                txen=eth.txen,
                txd0=eth.txd0,
                txd1=eth.txd1,
                uart_tx=uart_tx,
            ),
            0x200: ethernet.UartDevice(
                ctx=eth_ctx,
                rx=uart_rx,
                tx=uart_tx,
                baud=std.uart.Baud.BAUD_115200,
            ),
        }
    )

    class AllRegisters(reg.RegisterDeviceXX):
        SW = reg.R[0](switches)
        LED = reg.W[4](leds)
        WAIT = Wait[8]()
        BUTTONS = reg.R[12](buttons_vec)
        ETH = ethernet.EthernetDevice[0x100](
            ctx=eth_ctx,
            crsdv=eth.crsdv,
            rxd0=eth.rxd0,
            rxd1=eth.rxd1,
            txen=eth.txen,
            txd0=eth.txd0,
            txd1=eth.txd1,
            uart_tx=uart_tx,
        )

        UART = ethernet.UartDevice[0x200](
            ctx=eth_ctx,
            rx=uart_rx,
            tx=uart_tx,
            baud=std.uart.Baud.BAUD_115200,
        )

        SEVENT_SEG = ethernet.HexDisplay[0x300](
            ctx=eth_ctx, anodes=seven_seg.anodes, cathodes=seven_seg.cathodes
        )

        RGB_A = RgbLed[0x400](ctx=eth_ctx, rgb=rgb_1)
        RGB_B = RgbLed[0x404](ctx=eth_ctx, rgb=rgb_2)

        SPI = SpiMaster[0x800](
            ctx=eth_ctx, spi=acc.spi, clk_period=std.MHz(2.5).period()
        )

    RegisterDevice(
        interconnect.reserve(0x10000000, 10),
        [
            PartialRegister(0x00, switches, read_only=True),
            PartialRegister(0x04, leds, write_only=True),
            WaitRegister(0x08),
            PartialRegister(0x0C, button_vec, read_only=True),
            ethernet.EthernetDevice(
                0x10,
                ctx=eth_ctx,
                crsdv=eth.crsdv,
                rxd0=eth.rxd0,
                rxd1=eth.rxd1,
                txen=eth.txen,
                txd0=eth.txd0,
                txd1=eth.txd1,
                uart_tx=uart_tx,
            ),
            ethernet.UartDevice(
                0x20,
                ctx=eth_ctx,
                rx=uart_rx,
                tx=uart_tx,
                baud=std.uart.Baud.BAUD_115200,
            ),
            ethernet.HexDisplay(
                0x100, ctx=eth_ctx, anodes=seven_seg.anodes, cathodes=seven_seg.cathodes
            ),
            ethernet.RgbLed(0x180, ctx=eth_ctx, rgb=rgb_1),
            ethernet.RgbLed(0x184, ctx=eth_ctx, rgb=rgb_2),
            ethernet.SpiMaster(
                0x200, ctx=eth_ctx, spi=acc.spi, clk_period=std.MHz(2.5).period()
            ),
        ],
    )


if False:
    ctrl: CtrlRegister[0]

    send_low: reg.SimpleRegister.Wr[
        4,
        "dasd asist sfiasrlasfljsdalfölasd asdflsladjoislfldsafldn welsd wejjoinvhhsdf fsds",
    ]

    send_high: reg.SimpleRegister.Wr[8]

    receive_low: reg.SimpleRegister.Rd[12]
    receive_high: reg.SimpleRegister.Rd[16]

    def __init__(
        self,
        ctx: std.SequentialContext,
        spi: std.spi.Spi,
        clk_period=std.Duration,
    ):
        master = std.spi.SpiMaster(ctx, spi, clk_period=clk_period)

        send_data = Signal[BitVector[64]](Null)
        receive_data = Signal[BitVector[64]](Null)
        ctrl_register = Signal[BitVector[32]](Null)

        transaction_active = std.SyncFlag()

        @ctx
        async def proc():
            nonlocal receive_data

            async with transaction_active:
                receive_data <<= await master.parallel_transaction(
                    send_data, ctrl_register[5:0].unsigned
                )

        SimpleRegister()
        super().__init__(
            [
                CtrlRegister(0x00),
                SimpleRegister(0x04, send_data[31:0], write_only=True),
                SimpleRegister(0x08, send_data[63:32], write_only=True),
                SimpleRegister(0x0C, receive_data[31:0], read_only=True),
                SimpleRegister(0x10, receive_data[63:32], read_only=True),
            ],
        )



from typing import Callable


def merge_dicts(*dicts: dict, on_conflict: Callable):
    result: dict = {}

    for inp_dict in dicts:
        conflict_keys = result.keys() & inp_dict.keys()

        # insert unique items from inp_dict into result
        result.update({key: inp_dict[key] for key in inp_dict.keys() - conflict_keys})

        # resolve conflicts and put new items into result
        result.update(
            {key: on_conflict(inp_dict[key], result[key]) for key in conflict_keys}
        )

    return result



class _Always:
    def __call__(self, expr, /):
        return expr

    def __enter__(self) -> None: ...
    def __exit__(self, type, value, traceback): ...

always = _Always()
"""
`cohdl.always` creates concurrent logic within an sequential context.

When called the argument expression is hoisted out of the generated VHDL process
into the concurrent scope. The return value can be used from within the sequential
context.

>>> @std.sequential(clk)
>>> async def example():
>>>     sig <<= cohdl.Signal[Unsigned[4]](0)
>>>     incremented_sig = cohdl.always(sig + 1)
>>>     await std.tick()
>>>     # incremented_sig always follows changes in sig
>>>     assert incremented_sig == 1
>>>     sig <<= 2
>>>     await std.tick()
>>>     # incremented_sig always follows changes in sig
>>>     assert incremented_sig == 2

To implement more than one concurrent statement, `cohdl.always` is passed
to a Python context manager (with statement).

>>> inp_a = Signal[BitVector[8]]()
>>> inp_b = Signal[BitVector[8]]()
>>>
>>> @std.sequential(clk)
>>> async def example()
>>>     with cohdl.always:
>>>         result_and = inp_a & inp_b
>>>         result_or = inp_a | inp_b
>>>
>>>     assert result_and == inp_a & inp_b, "this assertion always holds"
>>>     await some_operation()
>>>     assert result_or == inp_a | inp_b, "so does this one"
"""

def expr(expr, /):
    """
    `expr` is used to hand an entire expression to the `await` operator.
    This is only possible in synthesizable contexts.

    ---
    Example:
    wait until one of two Signals becomes true

    a, b = Signal[Bit](), Signal[Bit]()

    ---
    `await a | b`
    this does not work because the await operator has higher precedence than the bitor operator
    i.e. Python evaluates it as `(await a) | b`

    ---
    `await (a | b)`
    is not allowed because await acts on the result of a|b
    and not the expression itself

    ---
    `await expr(a | b)`
    this works, the execution continues once a|b evaluates to true
    """

def expr_fn(fn, /):
    """
    When a function is decorated with `expr_fn` each call acts as if it
    were wrapped in a call to `expr`. This has no effect except when
    the result is used as the argument of an await expression.
    """

def evaluated() -> bool:
    """
    This function returns True when evaluated in a synthesizable context.
    When called in normal python code False is returned.
    """

def static_assert(cond: bool, msg: str | None = None) -> None: ...



from cohdl._core import (
    Bit,
    BitVector,
    Signal,
    Null,
)


class BitwiseCrc:
    def __init__(self, poly: BitVector, initial_value=Null, invert_result=False):
        self._invert_result = invert_result
        self._poly = poly.bitvector
        self._initial_value = initial_value
        self._reg = Signal[BitVector[poly.width]](initial_value)

    def _calc_steps(self, prev: BitVector, *data: Bit):
        first, *rest = data

        cond = prev.msb() ^ first
        shifted = prev.lsb(rest=1) @ Bit(0)
        result = (shifted ^ self._poly) if cond else shifted

        if len(rest) == 0:
            return result
        else:
            return self._calc_steps(result, *rest)

    def clear(self):
        self._reg <<= self._initial_value

    def update(self, data: Bit):
        cond = self._reg.msb() ^ data
        shifted = self._reg.lsb(rest=1) @ Bit(0)
        self._reg <<= (shifted ^ self._poly) if cond else shifted

    def update_multiple(self, *data: Bit):
        self._reg <<= self._calc_steps(self._reg, *data)

    def result(self):
        if self._invert_result:
            return ~self._reg
        else:
            return self._reg.copy()



from __future__ import annotations

from cohdl import vhdl, Bit, Signal
from cohdl import vhdl as raw_vhdl
from ._context import sequential

from cohdl import pyeval, Attribute

#
# inline vhdl
#


class vhdl(raw_vhdl):
    def post_process(self, text: str):
        result = text.replace("<%", "{")
        return result.replace("%>", "}")


#
# attributes
#


class anyconst(Attribute, type=bool): ...


#
#
#


class ConfigFormal:
    disable_cover = True
    disable_assert = True
    disable_assume = True


#
#
#


def conv(inp):
    if isinstance(inp, Formal):
        return inp.write()
    else:
        return f"{vhdl:{inp!r}}"


def conv_seq(s):
    if len(s) == 0:
        return f"{vhdl:}"
    else:
        first, *rest = s
        return f"{vhdl:{first.write()}{conv_seq(rest)}}"


#
#
#


class Formal:
    def write(self): ...


class When(Formal):
    @pyeval
    def __init__(self, precond, postcond=None, immediate=True):
        self.precond = precond
        self.postcond = postcond
        self.immediate = immediate

    @pyeval
    def then(self, state, times=None):
        if self.postcond is None:
            self.immediate = True
            self.postcond = State(state, times)
        elif isinstance(self.postcond, Sequence):
            self.postcond.then(state, times)
        else:
            self.postcond = Sequence(self.postcond).then(state, times)
        return self

    @pyeval
    def next(self, state, times=None):
        if self.postcond is None:
            self.immediate = False
            self.postcond = State(state, times)
        elif isinstance(self.postcond, Sequence):
            self.postcond.next(state, times)
        else:
            self.postcond = Sequence(self.postcond).next(state, times)
        return self

    def write(self):
        if self.immediate:
            return f"{vhdl:({conv(self.precond)} |-> {conv(self.postcond)})}"
        else:
            return f"{vhdl:({conv(self.precond)} |=> {conv(self.postcond)})}"


class State(Formal):
    def __init__(self, cond, times=None, consecutive=True):
        self.cond = cond
        self.times = times
        self.consecutive = consecutive

    def write_times(self):
        start_char = "*" if self.consecutive else "="

        if self.times is None:
            return ""
        elif isinstance(self.times, tuple):
            if len(self.times) == 0:
                return f"{vhdl:[*]}"
            elif len(self.times) == 1:
                raise AssertionError("not implemented")
            elif len(self.times) == 2:
                return f"{vhdl:[{start_char}{self.times[0]} to {self.times[1]}]}"
            else:
                raise AssertionError("")
        else:
            return f"{vhdl:[->{self.times}]}"

    def write(self):
        return f"{vhdl:{conv(self.cond)}{self.write_times()}}"


class Next(State):
    def __init__(self, cond, times=None, consecutive=True):
        super().__init__(cond, times, consecutive)

    def write(self):
        return f"{vhdl: ; {conv(self.cond)}{self.write_times()}}"


class Then(State):
    def __init__(self, cond, times=None, consecutive=True):
        super().__init__(cond, times, consecutive)

    def write(self):
        return f"{vhdl: : {conv(self.cond)}{self.write_times()}}"


class Wait(State):
    def __init__(self, times=None):
        super().__init__(None, times)

    def write(self):
        return f"{vhdl: ; true {self.write_times()}}"


class Sequence(Formal):
    def __init__(self, start, *seq):
        if isinstance(start, Formal):
            self.start = start
        else:
            self.start = State(start)

        self.seq = [*seq]

    @pyeval
    def then(self, state, times=None, consecutive=True):
        self.seq.append(Then(state, times=times, consecutive=consecutive))
        return self

    @pyeval
    def next(self, state, times=None, consecutive=True):
        self.seq.append(Next(state, times=times, consecutive=consecutive))
        return self

    @pyeval
    def wait(self, times):
        self.seq.append(Wait(times))
        return self

    def write(self):
        return f"{vhdl:<%{conv(self.start)}{conv_seq(self.seq)}%>}"


def stable(arg):
    return f"{vhdl[bool]:stable({arg!r})}"


def prev(state, cnt=None):
    if cnt is None:
        return f"{vhdl[state.type]:prev({state!r})}"
    else:
        return f"{vhdl[state.type]:prev({state!r}, {cnt})}"


class PrevCache:
    caches = {}

    def __init__(self, signal):
        self.signal = signal
        self.cache = {}
        self.cache_exact = {}

    @pyeval
    def get(self, cnt):
        return self.cache.get(cnt, None)

    @pyeval
    def get_exact(self, cnt):
        return self.cache_exact.get(cnt, None)

    @pyeval
    def add(self, cnt, signal):
        self.cache[cnt] = signal

    @pyeval
    def add_exact(self, cnt, signal):
        self.cache_exact[cnt] = signal

    @staticmethod
    @pyeval
    def get_cache(signal) -> PrevCache:
        for key, val in PrevCache.caches.items():
            if key == id(signal):
                return val

        new_cache = PrevCache(signal)
        PrevCache.caches[id(signal)] = new_cache
        return new_cache


def all_prev(state, cnt: int, *, exact=False):
    cache = PrevCache.get_cache(state)

    cached = cache.get(cnt)

    if cached is None:
        if cnt == 0:
            cache.add(cnt, state)
        else:
            cache.add(cnt, all_prev(state, cnt - 1) & prev(state, cnt))

    if exact:
        cached_exact = cache.get_exact(cnt)

        if cached_exact is None:
            cache.add_exact(cnt, cache.get(cnt) & ~prev(state, cnt + 1))

        return cache.get_exact(cnt)
    else:
        return cache.get(cnt)


class Checker:
    def __init__(self, clk, reset=None, *, prefix: str | None = None):
        self._always_false = Signal[Bit](False, name="always_false")
        self._past_valid = Signal[Bit](False, name="past_exists")
        self._past_valid_after_reset = Signal[Bit](False, name="past_valid_after_reset")
        self._clk = clk
        self._reset = reset
        self._prefix = prefix

        @sequential(clk, reset)
        def proc():
            self._past_valid_after_reset <<= True

        @sequential(clk)
        def proc():
            self._past_valid <<= True

    @pyeval
    def _complete_label(self, label):
        if self._prefix is not None:
            return f"{self._prefix}{label}"
        return label

    def valid_since_reset(self, cnt: int | None = None, *, exact=False):
        if cnt is None:
            return self._past_valid_after_reset
        else:
            return all_prev(self._past_valid_after_reset, cnt, exact=exact)

    def valid_since_start(self, cnt: int | None = None, *, exact=False):
        if cnt is None:
            return self._past_valid
        else:
            return all_prev(self._past_valid, cnt, exact=exact)

    def always(
        self,
        label,
        cond,
        *,
        since_start=None,
        since_reset=None,
    ):
        lbl = self._complete_label(label)

        if since_start is None:
            start_guard = True
        else:
            start_guard = self.valid_since_start(since_start)

        if since_reset is None:
            reset_guard = True
        else:
            reset_guard = self.valid_since_reset(since_reset)

        if start_guard is True and reset_guard is True:
            f"{vhdl:{lbl} : assert always {conv(cond)};}"
        else:
            f"{vhdl:{lbl} : assert always {conv(When(start_guard and reset_guard, cond))};}"

    #
    #
    #

    def assume_always(
        self,
        label,
        cond,
        *,
        since_start=None,
        since_reset=None,
    ):
        lbl = self._complete_label(label)

        if since_start is None:
            start_guard = True
        else:
            start_guard = self.valid_since_start(since_start)

        if since_reset is None:
            reset_guard = True
        else:
            reset_guard = self.valid_since_reset(since_reset)

        if start_guard is True and reset_guard is True:
            f"{vhdl:{lbl} : assume always {conv(cond)};}"
        else:
            f"{vhdl:{label} : assume always {conv(When(start_guard and reset_guard, cond))};}"

    def assume_initial(self, label, cond):
        f"{vhdl:{self._complete_label(label)} : assume {conv(cond)};}"

    #
    #
    #

    def cover(self, label, cond):
        f"{vhdl:{self._complete_label(label)} : cover <% {conv(cond)} %>;}"

    #
    #
    #

    def dbg_show(self, label, signal):
        self.always(label, signal, valid_for=10)

    def fail_after(self, label, cnt):
        self.always(
            self._complete_label(label),
            When(self.valid_since_start(cnt - 1), self._always_false),
        )



from cohdl._core import Bit, BitVector, Signal

from ._context import SequentialContext

import enum

class Baud(enum.IntEnum):
    """
    Enumeration of common uart baud rates
    """

    BAUD_110 = 110
    BAUD_300 = 300
    BAUD_600 = 600
    BAUD_1200 = 1200
    BAUD_2400 = 2400
    BAUD_4800 = 4800
    BAUD_9600 = 9600
    BAUD_14400 = 14400
    BAUD_19200 = 19200
    BAUD_38400 = 38400
    BAUD_57600 = 57600
    BAUD_115200 = 115200
    BAUD_128000 = 128000

class Parity(enum.Enum):
    NONE = enum.auto()
    EVEN = enum.auto()
    ODD = enum.auto()

class UartSender:
    def __init__(
        self,
        ctx: SequentialContext,
        tx: Signal[Bit],
        baud: int | Baud,
        bits: int = 8,
        stop_bits: int = 1,
        parity: Parity = Parity.NONE,
    ):
        """
        Check std.Uart for a documentation of the parameters.
        """
    async def send(self, data: BitVector):
        """
        Wait until uart transmitter is ready and send
        one word of data
        """

class UartReceiver:
    def __init__(
        self,
        ctx: SequentialContext,
        rx: Signal[Bit],
        baud: int | Baud,
        bits: int = 8,
        stop_bits=1,
        debounce: float | None = None,
        parity: Parity = Parity.NONE,
    ):
        """
        Check std.Uart for a documentation of the parameters.
        """
    def data_valid(self) -> bool:
        """
        Returns true when the UartReceiver received new data
        since the last call to clear_data or receive.
        """
    def clear_data(self):
        """ """
    def data(self) -> BitVector:
        """
        Return the last received data word.
        Only allowed when `data_valid()` returns true.
        """
    async def receive(self) -> BitVector:
        """
        Wait until new data is received and return it.
        """

class Uart:
    def __init__(
        self,
        ctx: SequentialContext,
        rx: Signal[Bit],
        tx: Signal[Bit],
        baud: int | Baud,
        bits: int = 8,
        stop_bits=1,
        debounce: float | None = None,
        parity: Parity = Parity.NONE,
    ):
        """
        Initialized a sender and a receiver with the same properties.

        :ctx:       Source of clocks and reset of the Uart logic.
                      The clock must have a defined frequency.
        :rx:        Signal with outgoing uart data
        :tx:        Signal with incoming uart data
        :baud:      baud rate
        :bits:      number of bits per word
        :stop_bits: number of stop bits
        :debounce:  When set the rx signal is debounced with a low pass filter.
                      The value defines the duration for a state change of the output
                      relative to a singe bit. Should be less than 0.5.
        :parity:    parity mode
        """

        self.sender: UartSender
        self.receiver: UartReceiver



from dataclasses import dataclass
import inspect


@dataclass
class SourceLocation:
    file: str | None
    line: int | None = None
    function: str | None = None

    def relative(self, offset: int):
        if self.line is None:
            new_line = None
        else:
            new_line = self.line + offset - 1

        return SourceLocation(self.file, new_line, self.function)

    @staticmethod
    def from_function(fn):
        return SourceLocation(
            inspect.getfile(fn), inspect.getsourcelines(fn)[1], fn.__name__
        )

    def __str__(self):
        if self.function is not None:
            return f"{self.function}:{self.file}:{self.line}"
        else:
            return f"{self.file}:{self.line}"



import os

from cohdl._compiler.frontend import generate_internal_representation
from cohdl._compiler.backend import generate_vhdl


class VhdlCompiler:
    @classmethod
    def to_ir(cls, entity):
        return generate_internal_representation(entity)

    @classmethod
    def to_vhdl_library(cls, top_entity):
        ir = generate_internal_representation(top_entity)
        return generate_vhdl(ir)

    @classmethod
    def to_string(cls, top_entity):
        return str(cls.to_vhdl_library(top_entity).write())

    @classmethod
    def to_dir(cls, top_entity, target_dir, *, mkdir: bool = False) -> list[str]:
        if not os.path.exists(target_dir):
            if mkdir:
                os.mkdir(target_dir)
            else:
                raise AssertionError(f"target directory '{target_dir}' does not exist")

        return cls.to_vhdl_library(top_entity).write_dir(target_dir)



from . import _repr as repr


from ._repr import (
    AccessFlags,
    Block,
    Concurrent,
    Sequential,
    Expression,
    Statement,
    _State,
    UnaryOp,
    BinOp,
    Compare,
    All,
)



from __future__ import annotations

import enum
from dataclasses import dataclass
from typing import Generic, TypeVar, Self, Any

T = TypeVar("T")
TemplateArg = TypeVar("TemplateArg")

class _TemplateMode(enum.Enum):
    ROOT = enum.auto()
    DECLARATION = enum.auto()
    SPECIALIZED = enum.auto()

class _TemplateMeta:
    arg: Any | None
    mode: _TemplateMode
    argtype: type
    instances: dict[str, type[Template]]
    annotations: dict[str, Any]

class Template(Generic[TemplateArg]):
    """
    Helper class, that implements a basic template functionality
    similar to C++.

    Useful because many `std` functions/containers operate on type
    arguments.

    >>> # Example:
    >>> # a simple template class, that holds two values
    >>> # and checks, their types against the provided template arguments
    >>> class _PairArgs:
    >>>     def __init__(self, args: tuple[type, type]):
    >>>         self._type_first, self._type_second = args
    >>>
    >>>     def __eq__(self, other: _PairArgs):
    >>>         # TemplateArg must define __eq__ to allow templates
    >>>         # instantiated with equivalent arguments to map to the same
    >>>         # specialized template class
    >>>         return (
    >>>             self._type_first == other._type_first
    >>>             and self._type_second == other._type_second
    >>>         )
    >>>
    >>> class Pair(std.Template[_PairArgs]):
    >>>     def __init__(self, first, second):
    >>>         assert type(first) is self._template_arg_._type_first
    >>>         assert type(second) is self._template_arg_._type_second
    >>>         self.first = first
    >>>         self.second = second
    >>>
    >>> a = Pair[int, float](1, 1.23)
    >>> b = Pair[int, str](2, "hello")
    >>> c = Pair[int, float](3, 34.5)
    >>> # templates instantiated with the same arguments
    >>> # map to the same type
    >>> assert type(a) is not type(b)
    >>> assert type(a) is type(c)
    """

    _template_meta_: _TemplateMeta
    """
    this class member contains the result of instantiating `TemplateArg`
    with the arguments used to specialize the template
    """

    @classmethod
    def _template_deduce_(cls, *args, **kwargs) -> type[Template[TemplateArg]]:
        """
        Given a set of constructor arguments, this classmethod
        should return the type fitting the arguments.
        """

    @classmethod
    def _template_specialize_(cls):
        """
        Called at the and of each new template specialization.
        Use to customize the generated types similar to __init_subclass__
        """

    def __class_getitem__(cls: Self, args) -> Self: ...

class _TemplateArg:
    Type: type

    def __call__(self, cls: type[T]) -> type[T]:
        """
        Decorator that adds __init__, __hash__ and __eq__
        to a class so it can be used as an argument for std.Template.
        """

template_arg = _TemplateArg()



from cohdl._core import Bit, BitVector, Unsigned, Signal, coroutine_step

from ._prefix import prefix, name
from ._context import SequentialContext, Duration
from .utility import (
    SyncFlag,
    ClockDivider,
    OutShiftRegister,
    InShiftRegister,
    debounce,
    wait_for,
)

from ._core_utility import parity as std_parity

import enum


class Baud(enum.IntEnum):
    BAUD_110 = 110
    BAUD_300 = 300
    BAUD_600 = 600
    BAUD_1200 = 1200
    BAUD_2400 = 2400
    BAUD_4800 = 4800
    BAUD_9600 = 9600
    BAUD_14400 = 14400
    BAUD_19200 = 19200
    BAUD_38400 = 38400
    BAUD_57600 = 57600
    BAUD_115200 = 115200
    BAUD_128000 = 128000


class Parity(enum.Enum):
    NONE = enum.auto()
    EVEN = enum.auto()
    ODD = enum.auto()


class UartSender:
    def __init__(
        self,
        ctx: SequentialContext,
        tx: Bit,
        baud: int | Baud,
        bits: int = 8,
        stop_bits: int = 1,
        parity: Parity = Parity.NONE,
    ):
        assert parity is Parity.NONE, "support for parity not yet implemented"

        with prefix("uarttx"):
            self._flag = SyncFlag()
            self._data = Signal[BitVector[bits]]()

            ticks_per_bit = Duration.seconds(1 / baud).count_periods(
                ctx.clk().period(), allowed_delta=0.001
            )
            bit_clk = ClockDivider(ctx, ticks_per_bit)

            async def proc_impl():
                nonlocal tx
                tx <<= True

                async with self._flag:
                    tx <<= False
                    shift_reg = OutShiftRegister(self._data)

                    while not shift_reg.empty():
                        tx <<= shift_reg.shift()

                    tx <<= True
                    await wait_for(stop_bits)

            @ctx
            def proc():
                if bit_clk.rising():
                    coroutine_step(proc_impl())

    async def send(self, data):
        assert not self._flag.is_set()
        self._data <<= data
        self._flag.set()
        await self._flag.is_clear()


_debounce_fn = debounce


class UartReceiver:
    def __init__(
        self,
        ctx: SequentialContext,
        rx: Bit,
        baud: int | Baud,
        bits: int = 8,
        stop_bits=1,
        debounce: float | None = None,
        parity: Parity = Parity.NONE,
    ):
        assert parity is Parity.NONE, "support for parity not yet implemented"

        has_parity = parity not in (None, Parity.NONE)
        parity_bit_cnt = 1 if has_parity else 0

        with prefix("uartrx"):
            if has_parity:
                parity_bit = Signal[Bit](False, name=name("parity"))

            bit_time = Duration.seconds(1.0 / baud)

            ticks_per_bit = bit_time.count_periods(
                ctx.clk().period(), allowed_delta=0.001
            )

            bit_clk = ClockDivider(ctx, ticks_per_bit, require_enable=True)

            if debounce is None:
                self._rx = rx
            else:
                assert 0.0 < debounce < 1.0, "debounce must be in the range 0-1"
                debounce_ticks = int(ticks_per_bit * debounce)
                assert debounce_ticks > 0, "debounce parameter to small"

                self._rx = _debounce_fn(
                    ctx, rx, debounce_ticks, initial=True, allowed_delta=0.01
                )

            self._flag = SyncFlag()
            self._data = Signal[BitVector[bits]]()

            received = InShiftRegister(1 + bits + parity_bit_cnt + stop_bits)
            active = Signal[Bit](False)

            bit_counter = Signal[Unsigned.upto(ticks_per_bit)](0)

            @ctx
            def proc_counter():
                if active:
                    if bit_counter == 0:
                        bit_counter.next = ticks_per_bit
                    else:
                        bit_counter.next = bit_counter - 1
                else:
                    bit_counter.next = 0

            @ctx
            def proc_clkgen():
                if not active:
                    if not self._rx:
                        bit_clk.enable()
                        active.next = True
                else:
                    if received.full():
                        bit_clk.disable()
                        active.next = False

                        if parity is Parity.NONE:
                            self._data <<= (
                                received.data().msb(rest=1).lsb(rest=stop_bits)
                            )
                            self._flag.set()
                        else:
                            p = std_parity(
                                received.data().msb(rest=1).lsb(rest=stop_bits)
                            )
                            data = (
                                received.data()
                                .msb(rest=1)
                                .lsb(rest=stop_bits + parity_bit_cnt)
                            )

                            if parity is Parity.EVEN:
                                if not p:
                                    self._data <<= data
                                    self._flag.set()
                            elif parity is Parity.ODD:
                                if p:
                                    self._data <<= data
                                    self._flag.set()

            @ctx
            def proc_receiver():
                if active:
                    if bit_counter == ticks_per_bit // 2:
                        received.shift(self._rx)
                else:
                    received.clear()

    def data_valid(self):
        return self._flag.is_set()

    def clear_data(self):
        self._flag.clear()

    def data(self):
        return self._data

    async def receive(self):
        async with self._flag:
            return self._data.copy()


class Uart:
    def __init__(
        self,
        ctx: SequentialContext,
        rx: Bit,
        tx: Bit,
        baud: int | Baud,
        bits: int = 8,
        stop_bits=1,
        debounce: float | None = None,
        parity: Parity = Parity.NONE,
    ):
        self.receiver = UartReceiver(
            ctx,
            rx=rx,
            baud=baud,
            bits=bits,
            stop_bits=stop_bits,
            debounce=debounce,
            parity=parity,
        )

        self.sender = UartSender(
            ctx, tx=tx, baud=baud, bits=bits, stop_bits=stop_bits, parity=parity
        )
